from math import pi, sqrt
import numpy as np
from ase.parallel import paropen
from gpaw.utilities import pack
from gpaw.analyse.wignerseitz import wignerseitz
from gpaw.setup_data import SetupData
from gpaw.gauss import Gauss
from gpaw.io.fmf import FMF
from gpaw.utilities.blas import gemmdot


def print_projectors(setup):
    """Print information on the projectors of input nucleus object.

    If nucleus is a string, treat this as an element name.
    """
    if isinstance(setup, str):
        setup = SetupData(setup, 'LDA', 'paw')
        n_j = setup.n_j
        l_j = setup.l_j
    else:
        n_j = setup.n_j
        l_j = setup.l_j

    angular = [['1'],
               ['y', 'z', 'x'],
               ['xy', 'yz', '3z^2-r^2', 'xz', 'x^2-y^2'],
               ['3x^2y-y^3', 'xyz', '5yz^2-yr^2', '5z^3-3zr^2',
                '5xz^2-xr^2', 'x^2z-y^2z', 'x^3-3xy^2'],
               ]
    print(' i n l m')
    print('--------')
    i = 0
    for n, l in zip(n_j, l_j):
        for m in range(2 * l + 1):
            if n == -1:
                n = '*'
            print('%2s %s %s_%s' % (i, n, 'spdf'[l], angular[l][m]))
            i += 1


def number_of_projectors(setup):
    """Returns the number of the bound state projectors.

    If setup is a string, treat this as an element name.
    """
    if isinstance(setup, str):
        setup = SetupData(setup, 'LDA', 'paw')
        n_j = setup.n_j
        l_j = setup.l_j
    else:
        n_j = setup.n_j
        l_j = setup.l_j

    i = 0
    for n, l in zip(n_j, l_j):
        for m in range(2 * l + 1):
            if n != -1:
                i += 1
    return i


def get_angular_projectors(setup, angular, type='bound'):
    """Determine the projector indices which have specified angula
    quantum number.

    angular can be s, p, d, f, or a list of these.
    If type is 'bound', only bound state projectors are considered, otherwise
    all projectors are included.
    """
    # Get the number of relevant j values
    if type == 'bound':
        nj = len([n for n in setup.n_j if n >= 0])
    else:
        nj = len(setup.n_j)

    # Choose the relevant projectors
    projectors = []
    i = 0
    j = 0
    for j in range(nj):
        m = 2 * setup.l_j[j] + 1
        if 'spdf'[setup.l_j[j]] in angular:
            projectors.extend(range(i, i + m))
        j += 1
        i += m

    return projectors


def delta(x, x0, width, mode='Gauss'):
    """Return a gaussian of given width centered at x0."""
    if mode == 'Gauss':
        return np.exp(np.clip(-((x - x0) / width)**2,
                              -100.0, 100.0)) / (sqrt(pi) * width)
    if mode == 'Lorentz':
        return (2 / pi / width) / ((np.clip(((x - x0) / (width / 2))**2,
                                            -100.0, 100.0)) + 1)


def fold(energies, weights, npts, width, mode='Gauss'):
    """Take a list of energies and weights, and sum a delta function
    for each."""
    emin = min(energies) - 5 * width
    emax = max(energies) + 5 * width
    e = np.linspace(emin, emax, npts)
    dos_e = np.zeros(npts)
    for e0, w in zip(energies, weights):
        dos_e += w * delta(e, e0, width, mode=mode)
    return e, dos_e


def raw_orbital_LDOS(paw, a, spin, angular='spdf', nbands=None):
    """Return a list of eigenvalues, and their weight on the specified atom.

    angular can be s, p, d, f, or a list of these.
    If angular is None, the raw weight for each projector is returned.

    An integer value for ``angular`` can also be used to specify a specific
    projector function.

    Setting nbands limits the number of bands included. This speeds up the
    calculation if one has many bands in the calculator but is only interested
    in the DOS at low energies.
    """
    wfs = paw.wfs
    w_k = wfs.kd.weight_k
    nk = len(w_k)
    if not nbands:
        nb = wfs.bd.nbands
    else:
        nb = nbands
        assert nb <= wfs.bd.nbands, ('nbands higher than available number' +
                                     'of bands')

    if a < 0:
        # Allow list-style negative indices; we'll need the positive a for the
        # dictionary lookup later
        a = len(wfs.setups) + a

    I1 = sum(setup.ni for setup in wfs.setups[:a])
    setup = wfs.setups[a]
    I2 = I1 + setup.ni

    energies = np.empty(nb * nk)
    weights_xi = np.empty((nb * nk, setup.ni))
    x = 0
    for k, w in enumerate(w_k):
        eps_n = wfs.collect_eigenvalues(k=k, s=spin)
        if len(eps_n) > 0:
            energies[x:x + nb] = eps_n[:nb]
        P_nI = wfs.collect_projections(k, spin)
        if P_nI is not None:
            weights_xi[x:x + nb, :] = w * abs(P_nI[:nb, I1:I2])**2
        x += nb

    wfs.world.broadcast(energies, 0)
    wfs.world.broadcast(weights_xi, 0)

    if angular is None:
        return energies, weights_xi
    elif isinstance(angular, int):
        return energies, weights_xi[:, angular]
    else:
        projectors = get_angular_projectors(setup, angular, type='bound')
        weights = np.sum(np.take(weights_xi,
                                 indices=projectors, axis=1), axis=1)
        return energies, weights


def all_electron_LDOS(paw, mol, spin, lc=None, wf_k=None, P_aui=None):
    """Returns a list of eigenvalues, and their weights on a given molecule

       wf_k should be a list of pseudo_wavefunctons of a Kohn-Sham state,
       corresponding to the different kpoints. It should be accompanied by a
       list of projector overlaps: P_aui=[[kpt.P_ani[a][n] for kpt in
       paw.wfs.kpt_u] for a in range(len(molecule))] for the band n. The
       weights are then calculated as the overlap of all-electron
       KS wavefunctions with wf_k

       If wf_k is None, the weights are calculated as linear combinations of
       atomic orbitals using P_uni. lc should then be a list of weights
       for each atom. For example, the pure 2pi_x orbital of a CO or N2
       molecule can be obtained with lc=[[0,0,0,1.0],[0,0,0,-1.0]]. mol
       should be a list of atom numbers contributing to the molecule."""

    w_k = paw.wfs.kd.weight_k
    nk = len(w_k)
    nb = paw.wfs.bd.nbands

    P_kn = np.zeros((nk, nb), complex)
    if wf_k is None:
        if lc is None:
            lc = [[1, 0, 0, 0] for a in mol]
        for k, kpt in enumerate(kpt_s[spin] for kpt_s in paw.wfs.kpt_qs):
            N = 0
            for atom, w_a in zip(mol, lc):
                i = 0
                for w_o in w_a:
                    P_kn[k] += w_o * kpt.P_ani[atom][:, i]
                    N += abs(w_o)**2
                    i += 1
        P_kn /= sqrt(N)

    else:
        P_aui = [np.array(P_ui).conj() for P_ui in P_aui]
        for k, kpt in enumerate(kpt_s[spin] for kpt_s in paw.wfs.kpt_qs):
            for n in range(nb):
                P_kn[k][n] = paw.wfs.integrate(wf_k[k], kpt.psit_nG[n])
                for a, b in zip(mol, range(len(mol))):
                    atom = paw.wfs.setups[a]
                    p_i = kpt.P_ani[a][n]
                    for i in range(len(p_i)):
                        for j in range(len(p_i)):
                            P_kn[k][n] += (P_aui[b][spin * nk + k][i] *
                                           atom.dO_ii[i][j] * p_i[j])
            print('# k', k, ' Sum_m |<m|n>|^2 =', sum(abs(P_kn[k])**2))

    energies = np.empty(nb * nk)
    weights = np.empty(nb * nk)
    x = 0
    for k, w in enumerate(w_k):
        eps_n = paw.wfs.collect_eigenvalues(k=k, s=spin)
        if len(eps_n) > 0:
            energies[x:x + nb] = eps_n
            weights[x:x + nb] = w * abs(P_kn[k])**2
        x += nb

    return energies, weights


def get_all_electron_IPR(paw):
    density = paw.density
    wfs = paw.wfs
    n_G = wfs.gd.empty()
    n_g = density.finegd.empty()
    print()
    print('inverse participation function')
    print('-' * 35)
    print('%5s %5s %10s %10s' % ('k', 'band', 'eps', 'ipr'))
    print('-' * 35)
    for k, kpt in enumerate(paw.wfs.kpt_u):
        for n, (eps, psit_G) in enumerate(zip(kpt.eps_n, kpt.psit_nG)):
            n_G[:] = 0.0
            wfs.add_orbital_density(n_G, kpt, n)
            density.interpolator.apply(n_G, n_g)
            norm = density.finegd.integrate(n_g)
            n_g = n_g ** 2
            ipr = density.finegd.integrate(n_g)
            for a in kpt.P_ani:
                # Get xccorr for atom a
                setup = paw.density.setups[a]
                xccorr = setup.xc_correction

                # Get D_sp for atom a
                D_sp = np.array(wfs.get_orbital_density_matrix(a, kpt, n))

                # density a function of L and partial wave radial pair
                # density coefficient
                D_sLq = gemmdot(D_sp, xccorr.B_Lqp, trans='t')

                # Create pseudo/ae density iterators for integration
                n_iter = xccorr.expand_density(D_sLq, xccorr.n_qg, None)
                nt_iter = xccorr.expand_density(D_sLq, xccorr.nt_qg, None)

                # Take the spherical average of smooth and ae radial
                # xc potentials
                for n_sg, nt_sg, integrator in zip(n_iter,
                                                   nt_iter,
                                                   xccorr.get_integrator(
                                                       None)):
                    ipr += integrator.weight * np.sum((n_sg[0]**2 -
                                                       nt_sg[0]**2) *
                                                      xccorr.rgd.dv_g)
                    norm += integrator.weight * np.sum((n_sg[0] - nt_sg[0]) *
                                                       xccorr.rgd.dv_g)

            print('%5i %5i %10.5f %10.5f' % (k, n, eps, ipr / norm**2))
    print('-' * 35)


def raw_wignerseitz_LDOS(paw, a, spin):
    """Return a list of eigenvalues, and their weight on the specified atom.
       Only supports real (gamma-point only) wavefunctions."""
    wfs = paw.wfs
    assert wfs.dtype == float
    gd = wfs.gd
    atom_index = wignerseitz(gd, paw.atoms)

    w_k = wfs.kd.weight_k
    nk = len(w_k)
    nb = wfs.bd.nbands

    energies = np.empty(nb * nk)
    weights = np.empty(nb * nk)
    x = 0
    for k, w in enumerate(w_k):
        u = spin * nk + k
        energies[x:x + nb] = wfs.collect_eigenvalues(k=k, s=spin)
        for n, psit_G in enumerate(wfs.kpt_u[u].psit_nG):
            P_i = wfs.kpt_u[u].P_ani[a][n]
            P_p = pack(np.outer(P_i, P_i))
            Delta_p = sqrt(4 * pi) * wfs.setups[a].Delta_pL[:, 0]
            weights[x + n] = w * (
                gd.integrate(abs(np.where(atom_index == a, psit_G, 0.0))**2) +
                np.dot(Delta_p, P_p))
        x += nb
    return energies, weights


class RawLDOS:
    """Class to get the unfolded LDOS"""
    def __init__(self, calc):
        self.paw = calc
        for setup in calc.wfs.setups.setups.values():
            if not hasattr(setup, 'l_i'):
                # get the mapping
                l_i = []
                for l in setup.l_j:
                    for m in range(2 * l + 1):
                        l_i.append(l)
                setup.l_i = l_i

    def get(self, atom):
        """Return the s,p,d weights for each state"""
        wfs = self.paw.wfs
        nibzkpts = len(wfs.kd.ibzk_kc)
        spd = np.zeros((wfs.nspins, nibzkpts, wfs.bd.nbands, 3))

        if hasattr(atom, '__iter__'):
            # atom is a list of atom indices
            for a in atom:
                spd += self.get(a)
            return spd

        l_i = wfs.setups[atom].l_i

        for kpt in self.paw.wfs.kpt_u:
            if atom in kpt.P_ani:
                for i, P_n in enumerate(kpt.P_ani[atom].T):
                    spd[kpt.s, kpt.k, :, l_i[i]] += np.abs(P_n)**2

        wfs.gd.comm.sum(spd)
        wfs.kd.comm.sum(spd)
        return spd

    def by_element(self, indices=None):
        """Return a dict with elements as keys and LDOS as values.

        Parameter:
        -----------
        indices: list or None
          Atom indices to consider, default None
        """
        atoms = self.paw.atoms
        if indices is None:
            selected = range(len(atoms))
        else:
            selected = indices

        elemi = {}
        for i in selected:
            symbol = atoms[i].symbol
            if symbol in elemi:
                elemi[symbol].append(i)
            else:
                elemi[symbol] = [i]
        for key in elemi:
            elemi[key] = self.get(elemi[key])
        return elemi

    def to_file(self,
                ldbe,
                filename,
                width=None,
                shift=True,
                bound=False):
        """Write the LDOS to a file.

        Parameters:
        -----------
        ldbe: dict
          weights
        filename: string
        width: float or None
          If a width is given, the LDOS will be Gaussian folded
        shift: bool
          Set Fermi energy to 0 eV. Default True
        bound: bool
          If the calculation was performed spin-polarized with fixmagmom=True,
          there are two Fermi-levels, one for each spin. False shifts
          both spins by their own Fermi levels, True both
          by the higher Fermi level. Default: False
        """

        f = paropen(filename, 'w')

        def append_weight_strings(ldbe, data):
            s = ''
            for key in ldbe:
                for l in 'spd':
                    data.append(key + '(' + l + ')-weight')
                if len(key) == 1:
                    key = ' ' + key
                s += ' ' + key + ':s     p        d       '
            return s

        wfs = self.paw.wfs

        if width is None:
            # unfolded ldos
            fmf = FMF(['Raw LDOS obtained from projector weights'])
            print(fmf.header(), end=' ', file=f)
            data = ['energy: energy [eV]',
                    'occupation number: occ',
                    'spin index: s',
                    'k-point index: k',
                    'band index: n',
                    'k-point weight: weight']
            string = '# e_i[eV]  occ     s     k      n   kptwght '
            string += append_weight_strings(ldbe, data)
            data.append('summed weights: sum')
            string += ' sum'
            print(fmf.data(data), end=' ', file=f)
            print(string, file=f)
            for k in range(wfs.kd.nibzkpts):
                for s in range(wfs.nspins):
                    e_n = self.paw.get_eigenvalues(kpt=k, spin=s)
                    f_n = self.paw.get_occupation_numbers(kpt=k, spin=s)
                    if e_n is None:
                        continue
                    w = wfs.kd.weight_k[k]
                    for n in range(wfs.bd.nbands):
                        sum = 0.0
                        print('%10.5f %6.4f %2d %5d' % (e_n[n], f_n[n],
                                                        s, k), end=' ', file=f)
                        print('%6d %8.4f' % (n, w), end=' ', file=f)
                        for key in ldbe:
                            spd = ldbe[key][s, k, n]
                            for l in range(3):
                                sum += spd[l]
                                print('%8.4f' % spd[l], end=' ', file=f)
                        print('%8.4f' % sum, file=f)
        else:
            # folded ldos
            fmf = FMF(['Folded raw LDOS obtained from projector weights'])
            print(fmf.header(), end=' ', file=f)

            gauss = Gauss(width)
            print(fmf.field('folding',
                            ['function: Gauss',
                             'width: ' + str(width) + ' [eV]']),
                  end=' ', file=f)

            data = ['energy: energy [eV]',
                    'spin index: s',
                    'k-point index: k',
                    'band index: n',
                    'k-point weight: weight']

            # minimal and maximal energies
            emin = 1.e32
            emax = -1.e32
            for k in range(wfs.kd.nibzkpts):
                for s in range(wfs.nspins):
                    e_n = self.paw.get_eigenvalues(kpt=k, spin=s)
                    emin = min(e_n.min(), emin)
                    emax = max(e_n.max(), emax)
            emin -= 4 * width
            emax += 4 * width

            # Fermi energies
            efermis = self.paw.wfs.fermi_levels

            eshift = 0.0

            if shift and len(efermis) == 1:
                eshift = -efermis[0]

            # set de to sample 4 points in the width
            de = width / 4

            string = '# e[eV]     s  '
            string += append_weight_strings(ldbe, data)

            for s in range(wfs.nspins):
                if len(efermis) == 2:
                    if not bound:
                        eshift = -efermis[s]
                    else:
                        eshift = -efermis.max()

                print(fmf.data(data), end=' ', file=f)

                print('# Gauss folded, width=%g [eV]' % width, file=f)
                if shift:
                    print('# shifted to Fermi energy = 0', file=f)
                    print('# Fermi energy was', end=' ', file=f)
                else:
                    print('# Fermi energy', end=' ', file=f)
                print(efermis, 'eV', file=f)
                print(string, file=f)

                # loop over energies
                emax = emax + 0.5 * de
                e = emin
                while e < emax:
                    val = {}
                    for key in ldbe:
                        val[key] = np.zeros((3))
                    for k in range(wfs.kd.nibzkpts):
                        w = wfs.kpt_u[k].weight
                        e_n = self.paw.get_eigenvalues(kpt=k, spin=s)
                        for n in range(wfs.bd.nbands):
                            w_i = w * gauss.get(e_n[n] - e)
                            for key in ldbe:
                                val[key] += w_i * ldbe[key][s, k, n]

                    print('%10.5f %2d' % (e + eshift, s), end=' ', file=f)
                    for key in val:
                        spd = val[key]
                        for l in range(3):
                            print('%8.4f' % spd[l], end=' ', file=f)
                    print(file=f)
                    e += de

        f.close()

    def by_element_to_file(self,
                           filename='ldos_by_element.dat',
                           width=None,
                           shift=True,
                           bound=False,
                           indices=None):
        """
        Write the LDOS by element to a file.

        Parameters:
        -----------
        indices: list or None
          Atom indices to consider, default None
        """
        ldbe = self.by_element(indices)
        self.to_file(ldbe, filename, width, shift, bound)


class LCAODOS:
    """Class for calculating the projected subspace DOS.

    The projected subspace density of states is defined only in LCAO
    mode.  The advantages to the PDOS based on projectors is that the
    LCAODOS will properly take non-orthogonality and completeness into
    account."""
    def __init__(self, calc):
        self.calc = calc

    def get_orbital_pdos(self, M, ravel=True):
        """Get projected DOS from LCAO basis function M."""
        return self.get_subspace_pdos([M], ravel=ravel)

    def get_atomic_subspace_pdos(self, a, ravel=True):
        """Get projected subspace DOS from LCAO basis on atom(s) a.

        a may be an atomic index or a list of indices."""
        Mvalues = self.get_atom_indices(a)
        return self.get_subspace_pdos(Mvalues, ravel=ravel)

    def get_atom_indices(self, a):
        """Get list of basis function indices of atom(s) a."""
        if isinstance(a, int):
            a = [a]
        Mvalues = []
        for a0 in a:
            M = self.calc.wfs.basis_functions.M_a[a0]
            Mvalues.extend(range(M, M + self.calc.wfs.setups[a0].nao))
        return Mvalues

    def get_subspace_pdos(self, Mvalues, ravel=True):
        """Get projected subspace DOS from LCAO basis."""
        wfs = self.calc.wfs
        bd = wfs.bd
        kd = wfs.kd
        gd = wfs.gd

        for kpt in wfs.kpt_u:
            assert not np.isnan(kpt.eps_n).any()

        w_skn = np.zeros((kd.nspins, kd.nibzkpts, bd.nbands))
        eps_skn = np.zeros((kd.nspins, kd.nibzkpts, bd.nbands))
        for u, kpt in enumerate(wfs.kpt_u):
            C_nM = kpt.C_nM
            from gpaw.kohnsham_layouts import BlacsOrbitalLayouts
            if isinstance(wfs.ksl, BlacsOrbitalLayouts):
                raise NotImplementedError('Something not quite working.  '
                                          'FIXME.')
                S_MM = wfs.ksl.mmdescriptor.collect_on_master(kpt.S_MM)
                if bd.rank != 0 or gd.rank != 0:
                    S_MM = np.empty((wfs.ksl.nao, wfs.ksl.nao),
                                    dtype=wfs.dtype)
                bd.comm.broadcast(S_MM, 0)
            else:
                S_MM = kpt.S_MM

            if gd.comm.rank == 0:
                S_mm = S_MM[Mvalues, :][:, Mvalues].copy()
                iS_mm = np.linalg.inv(S_mm).copy()
                CS_nm = np.dot(C_nM, S_MM[:, Mvalues])
                CSiS_nm = np.dot(CS_nm, iS_mm)
                w_n = (np.conj(CS_nm) * CSiS_nm).sum(axis=1) * kpt.weight
                w_skn[kpt.s, kpt.k, bd.beg:bd.end:bd.step] = w_n.real
                eps_skn[kpt.s, kpt.k, bd.beg:bd.end:bd.step] = kpt.eps_n

        for arr in [eps_skn, w_skn]:
            gd.comm.broadcast(arr, 0)
            bd.comm.sum(arr)
            kd.comm.sum(arr)

        if ravel:
            eps_n = eps_skn.ravel()
            w_n = w_skn.ravel()
            args = np.argsort(eps_n)
            eps_n = eps_n[args]
            w_n = w_n[args]
            return eps_n, w_n
        else:
            return eps_skn, w_skn


class RestartLCAODOS(LCAODOS):
    """Class for calculating LCAO subspace PDOS from a restarted calculator.

    Warning: This has side effects on the calculator.  The
    operation will allocate memory to diagonalize the Hamiltonian and
    set coefficients plus positions."""
    def __init__(self, calc):
        LCAODOS.__init__(self, calc)
        system = calc.get_atoms()
        calc.set_positions(system)
        calc.wfs.eigensolver.iterate(calc.hamiltonian, calc.wfs)



# Test loading of sg15 setups as setups='sg15' and that the calculation
# agrees with PAW for the H2 eigenvalue.

from ase.build import molecule
from gpaw import GPAW, Davidson, Mixer


def test_pseudopotential_sg15_hydrogen(sg15_hydrogen):
    system = molecule('H2')
    system.center(vacuum=2.5)

    def getkwargs():
        return dict(mode='fd',
                    eigensolver=Davidson(4),
                    mixer=Mixer(0.8, 5, 10.0),
                    xc='oldPBE')

    calc1 = GPAW(setups={'H': sg15_hydrogen}, h=0.13, **getkwargs())
    system.calc = calc1
    system.get_potential_energy()
    eps1 = calc1.get_eigenvalues()

    calc2 = GPAW(h=0.2, **getkwargs())
    system.calc = calc2
    system.get_potential_energy()
    eps2 = calc2.get_eigenvalues()

    err = eps2[0] - eps1[0]

    # It is not the most accurate calculation ever, let's just make sure things
    # are not completely messed up.
    print('sg15 vs paw error', err)
    assert abs(err) < 0.02  # 0.0055.... as of current test.



from __future__ import annotations

import numpy as np

from ase.units import Hartree

from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.pw_parallelization import Blocks1D
from gpaw.response.pair_functions import (SingleQPWDescriptor, Chi,
                                          get_pw_coordinates)
from gpaw.response.chiks import ChiKSCalculator
from gpaw.response.coulomb_kernels import get_coulomb_kernel
from gpaw.response.fxc_kernels import FXCKernel, AdiabaticFXCCalculator
from gpaw.response.dyson import DysonSolver, HXCKernel


class ChiFactory:
    r"""User interface to calculate individual elements of the four-component
    susceptibility tensor χ^μν, see [PRB 103, 245110 (2021)]."""

    def __init__(self,
                 chiks_calc: ChiKSCalculator,
                 fxc_calculator: AdiabaticFXCCalculator | None = None):
        """Contruct a many-body susceptibility factory."""
        self.chiks_calc = chiks_calc
        self.gs = chiks_calc.gs
        self.context = chiks_calc.context
        self.dyson_solver = DysonSolver(self.context)

        # If no fxc_calculator is supplied, fall back to default
        if fxc_calculator is None:
            fxc_calculator = AdiabaticFXCCalculator.from_rshe_parameters(
                self.gs, self.context)
        else:
            assert fxc_calculator.gs is chiks_calc.gs
            assert fxc_calculator.context is chiks_calc.context
        self.fxc_calculator = fxc_calculator

        # Prepare a buffer for the fxc kernels
        self.fxc_kernel_cache: dict[str, FXCKernel] = {}

    def __call__(self, spincomponent, q_c, complex_frequencies,
                 fxc=None, hxc_scaling=None, txt=None) -> tuple[Chi, Chi]:
        r"""Calculate a given element (spincomponent) of the four-component
        Kohn-Sham susceptibility tensor and construct a corresponding many-body
        susceptibility object within a given approximation to the
        exchange-correlation kernel.

        Parameters
        ----------
        spincomponent : str
            Spin component (μν) of the susceptibility.
            Currently, '00', 'uu', 'dd', '+-' and '-+' are implemented.
        q_c : list or ndarray
            Wave vector
        complex_frequencies : np.array or ComplexFrequencyDescriptor
            Array of complex frequencies to evaluate the response function at
            or a descriptor of those frequencies.
        fxc : str (None defaults to ALDA)
            Approximation to the (local) xc kernel.
            Choices: RPA, ALDA, ALDA_X, ALDA_x
        hxc_scaling : None or HXCScaling
            Supply an HXCScaling object to scale the hxc kernel.
        txt : str
            Save output of the calculation of this specific component into
            a file with the filename of the given input.
        """
        # Fall back to ALDA per default
        if fxc is None:
            fxc = 'ALDA'

        # Initiate new output file, if supplied
        if txt is not None:
            self.context.new_txt_and_timer(txt)

        # Print to output file
        self.context.print('---------------', flush=False)
        self.context.print('Calculating susceptibility spincomponent='
                           f'{spincomponent} with q_c={q_c}', flush=False)
        self.context.print('---------------')

        # Calculate chiks
        chiks = self.calculate_chiks(spincomponent, q_c, complex_frequencies)

        # Construct the hxc kernel
        hartree_kernel = self.get_hartree_kernel(spincomponent, chiks.qpd)
        xc_kernel = self.get_xc_kernel(fxc, spincomponent, chiks.qpd)
        hxc_kernel = HXCKernel(hartree_kernel, xc_kernel, scaling=hxc_scaling)

        # Solve dyson equation
        chi = self.dyson_solver(chiks, hxc_kernel)

        return chiks, chi

    def get_hartree_kernel(self, spincomponent, qpd):
        if spincomponent in ['+-', '-+']:
            # No Hartree term in Dyson equation
            return None
        else:
            return get_coulomb_kernel(qpd, self.gs.kd.N_c,
                                      pbc_c=self.gs.atoms.get_pbc())

    def get_xc_kernel(self,
                      fxc: str,
                      spincomponent: str,
                      qpd: SingleQPWDescriptor):
        """Get the requested xc-kernel object."""
        if fxc == 'RPA':
            # No xc-kernel
            return None

        if qpd.gammacentered:
            # When using a gamma-centered plane-wave basis, we can reuse the
            # fxc kernel for all q-vectors. Thus, we keep a cache of calculated
            # kernels
            key = f'{fxc},{spincomponent}'
            if key not in self.fxc_kernel_cache:
                self.fxc_kernel_cache[key] = self.fxc_calculator(
                    fxc, spincomponent, qpd)
            fxc_kernel = self.fxc_kernel_cache[key]
        else:
            # Always compute the kernel
            fxc_kernel = self.fxc_calculator(fxc, spincomponent, qpd)

        return fxc_kernel

    def calculate_chiks(self, spincomponent, q_c, complex_frequencies):
        """Calculate the Kohn-Sham susceptibility."""
        q_c = np.asarray(q_c)
        if isinstance(complex_frequencies, ComplexFrequencyDescriptor):
            zd = complex_frequencies
        else:
            zd = ComplexFrequencyDescriptor.from_array(complex_frequencies)

        # Perform actual calculation
        chiks = self.chiks_calc.calculate(spincomponent, q_c, zd)
        # Distribute frequencies over world
        chiks = chiks.copy_with_global_frequency_distribution()

        return chiks


def spectral_decomposition(chi, pos_eigs=1, neg_eigs=0):
    """Decompose the susceptibility in terms of spectral functions.

    The full spectrum of induced excitations is extracted and separated into
    contributions corresponding to the pos_eigs and neg_eigs largest positive
    and negative eigenvalues respectively.
    """
    # Initiate an EigendecomposedSpectrum object with the full spectrum
    full_spectrum = EigendecomposedSpectrum.from_chi(chi)

    # Separate the positive and negative eigenvalues for each frequency
    Apos = full_spectrum.get_positive_eigenvalue_spectrum()
    Aneg = full_spectrum.get_negative_eigenvalue_spectrum()

    # Keep only a fixed number of eigenvalues
    Apos = Apos.reduce_number_of_eigenvalues(pos_eigs)
    Aneg = Aneg.reduce_number_of_eigenvalues(neg_eigs)

    return Apos, Aneg


class EigendecomposedSpectrum:
    """Data object for eigendecomposed susceptibility spectra."""

    def __init__(self, omega_w, G_Gc, s_we, v_wGe, A_w=None,
                 wblocks: Blocks1D | None = None):
        """Construct the EigendecomposedSpectrum.

        Parameters
        ----------
        omega_w : np.array
            Global array of frequencies in eV.
        G_Gc : np.array
            Reciprocal lattice vectors in relative coordinates.
        s_we : np.array
            Sorted eigenvalues (in decreasing order) at all frequencies.
            Here, e is the eigenvalue index.
        v_wGe : np.array
            Eigenvectors for corresponding to the (sorted) eigenvalues. With
            all eigenvalues present in the representation, v_Ge should
            constitute the unitary transformation matrix between the eigenbasis
            and the plane-wave representation.
        A_w : np.array or None
            Full spectral weight as a function of frequency. If given as None,
            A_w will be calculated as the sum of all eigenvalues (equal to the
            trace of the spectrum, if no eigenvalues have been discarded).
        wblocks : Blocks1D
            Frequency block parallelization, if any.
        """
        self.omega_w = omega_w
        self.G_Gc = G_Gc

        self.s_we = s_we
        self.v_wGe = v_wGe

        self._A_w = A_w
        if wblocks is None:
            # Create a serial Blocks1D instance
            from gpaw.mpi import serial_comm
            wblocks = Blocks1D(serial_comm, len(omega_w))
        self.wblocks = wblocks

    @classmethod
    def from_chi(cls, chi):
        """Construct the eigendecomposed spectrum of a given susceptibility.

        The spectrum of induced excitations, S_GG'^(μν)(q,ω), which are encoded
        in a given susceptibility, can be extracted directly from its the
        dissipative part:

                            1
        S_GG'^(μν)(q,ω) = - ‾ χ_GG'^(μν")(q,ω)
                            π
        """
        assert chi.distribution == 'zGG'

        # Extract the spectrum of induced excitations
        chid = chi.copy_dissipative_part()
        S_wGG = - chid.array / np.pi

        # Extract frequencies (in eV) and reciprocal lattice vectors
        omega_w = chid.zd.omega_w * Hartree
        G_Gc = get_pw_coordinates(chid.qpd)

        return cls.from_spectrum(omega_w, G_Gc, S_wGG, wblocks=chid.blocks1d)

    @classmethod
    def from_spectrum(cls, omega_w, G_Gc, S_wGG, wblocks=None):
        """Perform an eigenvalue decomposition of a given spectrum."""
        # Find eigenvalues and eigenvectors of the spectrum
        s_wK, v_wGK = np.linalg.eigh(S_wGG)

        # Sort by spectral intensity (eigenvalues in descending order)
        sorted_indices_wK = np.argsort(-s_wK)
        s_we = np.take_along_axis(s_wK, sorted_indices_wK, axis=1)
        v_wGe = np.take_along_axis(
            v_wGK, sorted_indices_wK[:, np.newaxis, :], axis=2)

        return cls(omega_w, G_Gc, s_we, v_wGe, wblocks=wblocks)

    @classmethod
    def from_file(cls, filename):
        """Construct the eigendecomposed spectrum from a .pckl file."""
        import pickle
        assert isinstance(filename, str) and filename[-5:] == '.pckl'
        with open(filename, 'rb') as fd:
            omega_w, G_Gc, s_we, v_wGe, A_w = pickle.load(fd)
        return cls(omega_w, G_Gc, s_we, v_wGe, A_w=A_w)

    def write(self, filename):
        """Write the eigendecomposed spectrum as a .pckl file."""
        import pickle
        assert isinstance(filename, str) and filename[-5:] == '.pckl'

        # Gather data from the different blocks of frequencies to root
        s_we = self.wblocks.gather(self.s_we)
        v_wGe = self.wblocks.gather(self.v_wGe)
        A_w = self.wblocks.gather(self.A_w)

        # Let root write the spectrum to a pickle file
        if self.wblocks.blockcomm.rank == 0:
            with open(filename, 'wb') as fd:
                pickle.dump((self.omega_w, self.G_Gc, s_we, v_wGe, A_w), fd)

    @property
    def nG(self):
        return self.G_Gc.shape[0]

    @property
    def neigs(self):
        return self.s_we.shape[1]

    @property
    def A_w(self):
        if self._A_w is None:
            self._A_w = np.nansum(self.s_we, axis=1)
        return self._A_w

    @property
    def A_wGG(self):
        """Generate the spectrum from the eigenvalues and eigenvectors."""
        A_wGG = np.empty((self.wblocks.nlocal, self.nG, self.nG),
                         dtype=complex)
        for w, (s_e, v_Ge) in enumerate(zip(self.s_we, self.v_wGe)):
            emask = ~np.isnan(s_e)
            svinv_eG = s_e[emask][:, np.newaxis] * np.conj(v_Ge.T[emask])
            A_wGG[w] = v_Ge[:, emask] @ svinv_eG
        return A_wGG

    def get_positive_eigenvalue_spectrum(self):
        """Create a new EigendecomposedSpectrum from the positive eigenvalues.

        This is especially useful in order to separate the full spectrum of
        induced excitations, see [PRB 103, 245110 (2021)],

        S_GG'^μν(q,ω) = A_GG'^μν(q,ω) - A_(-G'-G)^νμ(-q,-ω)

        into the ν and μ components of the spectrum. Since the spectral
        function A_GG'^μν(q,ω) is positive definite or zero (in regions without
        excitations), A_GG'^μν(q,ω) simply corresponds to the positive
        eigenvalue contribution to the full spectrum S_GG'^μν(q,ω).
        """
        # Find the maximum number of positive eigenvalues across the entire
        # frequency range
        if self.wblocks.nlocal > 0:
            pos_we = self.s_we > 0.
            npos_max = int(np.max(np.sum(pos_we, axis=1)))
        else:
            npos_max = 0
        npos_max = self.wblocks.blockcomm.max(npos_max)

        # Allocate new arrays filled with nan to accomodate all the positive
        # eigenvalues
        s_we = np.empty((self.wblocks.nlocal, npos_max),
                        dtype=self.s_we.dtype)
        v_wGe = np.empty((self.wblocks.nlocal, self.nG, npos_max),
                         dtype=self.v_wGe.dtype)
        s_we[:] = np.nan
        v_wGe[:] = np.nan

        # Fill arrays with the positive eigenvalue data
        for w, (s_e, v_Ge) in enumerate(zip(self.s_we, self.v_wGe)):
            pos_e = s_e > 0.
            npos = np.sum(pos_e)
            s_we[w, :npos] = s_e[pos_e]
            v_wGe[w, :, :npos] = v_Ge[:, pos_e]

        return EigendecomposedSpectrum(self.omega_w, self.G_Gc, s_we, v_wGe,
                                       wblocks=self.wblocks)

    def get_negative_eigenvalue_spectrum(self):
        """Create a new EigendecomposedSpectrum from the negative eigenvalues.

        The spectrum is created by reversing and negating the spectrum,

        -S_GG'^μν(q,-ω) = -A_GG'^μν(q,-ω) + A_(-G'-G)^νμ(-q,ω),

        from which the spectral function A_GG'^νμ(q,ω) can be extracted as the
        positive eigenvalue contribution, thanks to the reciprocity relation

                                  ˍˍ
        χ_GG'^μν(q,ω) = χ_(-G'-G)^νμ(-q,ω),
                   ˍ
        in which n^μ(r) denotes the hermitian conjugate [n^μ(r)]^†, and which
        is valid for μν ∊ {00,0z,zz,+-} in collinear systems without spin-orbit
        coupling.
        """
        # Negate the spectral function, its frequencies and reverse the order
        # of eigenvalues
        omega_w = - self.omega_w
        s_we = - self.s_we[:, ::-1]
        v_wGe = self.v_wGe[..., ::-1]
        inverted_spectrum = EigendecomposedSpectrum(omega_w, self.G_Gc,
                                                    s_we, v_wGe,
                                                    wblocks=self.wblocks)

        return inverted_spectrum.get_positive_eigenvalue_spectrum()

    def reduce_number_of_eigenvalues(self, neigs):
        """Create a new spectrum with only the neigs largest eigenvalues.

        The returned EigendecomposedSpectrum is constructed to retain knowledge
        of the full spectral weight of the unreduced spectrum through the A_w
        attribute.
        """
        assert self.neigs >= neigs
        # Check that the available eigenvalues are in descending order
        assert all([np.all(np.logical_not(s_e[1:] - s_e[:-1] > 0.))
                    for s_e in self.s_we]), \
            'Eigenvalues needs to be sorted in descending order!'

        # Keep only the neigs largest eigenvalues
        s_we = self.s_we[:, :neigs]
        v_wGe = self.v_wGe[..., :neigs]

        return EigendecomposedSpectrum(self.omega_w, self.G_Gc, s_we, v_wGe,
                                       # Keep the full spectral weight
                                       A_w=self.A_w,
                                       wblocks=self.wblocks)

    def get_eigenmode_lineshapes(self, nmodes=1):
        """Extract the spectral lineshapes of the eigenmodes.

        The spectral lineshape is calculated as the inner product


        a^μν_n(q,ω) = <v^μν_n(q)| A^μν(q,ω) |v^μν_n(q)>

        where the eigenvectors |v^μν_n> diagonalize the full spectral function
        at an appropriately chosen frequency ω_m:

        S^μν(q,ω_m) |v^μν_n(q)> = s^μν_n(q,ω_m) |v^μν_n(q)>
        """
        wm = self.get_eigenmode_frequency(nmodes=nmodes)
        v_Gm = self.get_eigenvectors_at_frequency(wm, nmodes=nmodes)
        return self._get_eigenmode_lineshapes(v_Gm)

    def get_eigenmode_frequency(self, nmodes=1):
        """Get the frequency at which to extract the eigenmodes.

        Generally, we chose the frequency ω_m to maximize the minimum
        eigenvalue difference

        ω_m(q) = maxmin_ωn[s^μν_n(q,ω) - s^μν_n+1(q,ω)]

        where n only runs over the desired number of modes (and the eigenvalues
        are sorted in descending order).

        However, in the case where only a single mode is extracted, we use the
        frequency at which the eigenvalue is maximal.
        """
        assert nmodes <= self.neigs
        wblocks = self.wblocks
        if nmodes == 1:
            # Find frequency where the eigenvalue is maximal
            s_w = wblocks.all_gather(self.s_we[:, 0])
            wm = np.argmax(s_w)
        else:
            # Find frequency with maximum minimal difference between size of
            # eigenvalues
            ds_we = np.array([self.s_we[:, e] - self.s_we[:, e + 1]
                              for e in range(nmodes - 1)]).T
            dsmin_w = np.min(ds_we, axis=1)
            dsmin_w = wblocks.all_gather(dsmin_w)
            wm = np.argmax(dsmin_w)

        return wm

    def get_eigenvectors_at_frequency(self, wm, nmodes=1):
        """Extract the eigenvectors a specific frequency index."""
        wblocks = self.wblocks
        root, wmlocal = wblocks.find_global_index(wm)
        if wblocks.blockcomm.rank == root:
            v_Ge = self.v_wGe[wmlocal]
            v_Gm = np.ascontiguousarray(v_Ge[:, :nmodes])
        else:
            v_Gm = np.empty((self.nG, nmodes), dtype=complex)
        wblocks.blockcomm.broadcast(v_Gm, root)

        return v_Gm

    def _get_eigenmode_lineshapes(self, v_Gm):
        """Extract the eigenmode lineshape based on the mode eigenvectors."""
        wblocks = self.wblocks
        A_wGG = self.A_wGG
        a_wm = np.empty((wblocks.nlocal, v_Gm.shape[1]), dtype=float)
        for m, v_G in enumerate(v_Gm.T):
            a_w = np.conj(v_G) @ A_wGG @ v_G
            assert np.allclose(a_w.imag, 0.)
            a_wm[:, m] = a_w.real
        a_wm = wblocks.all_gather(a_wm)

        return a_wm

    def write_full_spectral_weight(self, filename):
        A_w = self.wblocks.gather(self.A_w)
        if self.wblocks.blockcomm.rank == 0:
            write_full_spectral_weight(filename, self.omega_w, A_w)

    def write_eigenmode_lineshapes(self, filename, nmodes=1):
        a_wm = self.get_eigenmode_lineshapes(nmodes=nmodes)
        if self.wblocks.blockcomm.rank == 0:
            write_eigenmode_lineshapes(filename, self.omega_w, a_wm)


def write_full_spectral_weight(filename, omega_w, A_w):
    """Write the sum of spectral weights A(ω) to a file."""
    with open(filename, 'w') as fd:
        print('# {0:>11}, {1:>11}'.format('omega [eV]', 'A(w)'), file=fd)
        for omega, A in zip(omega_w, A_w):
            print('  {0:11.6f}, {1:11.6f}'.format(omega, A), file=fd)


def read_full_spectral_weight(filename):
    """Read a stored full spectral weight file."""
    data = np.loadtxt(filename, delimiter=',')
    omega_w = np.array(data[:, 0], float)
    A_w = np.array(data[:, 1], float)
    return omega_w, A_w


def write_eigenmode_lineshapes(filename, omega_w, a_wm):
    """Write the eigenmode lineshapes a^μν_n(ω) to a file."""
    with open(filename, 'w') as fd:
        # Print header
        header = '# {0:>11}'.format('omega [eV]')
        for m in range(a_wm.shape[1]):
            header += ', {0:>11}'.format(f'a_{m}(w)')
        print(header, file=fd)
        # Print data
        for omega, a_m in zip(omega_w, a_wm):
            data = '  {0:11.6f}'.format(omega)
            for a in a_m:
                data += ', {0:11.6f}'.format(a)
            print(data, file=fd)


def read_eigenmode_lineshapes(filename):
    """Read a stored eigenmode lineshapes file."""
    data = np.loadtxt(filename, delimiter=',')
    omega_w = np.array(data[:, 0], float)
    a_wm = np.array(data[:, 1:], float)
    return omega_w, a_wm



from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.test import equal
from ase.build import molecule
from ase.units import mol, kcal
from gpaw.directmin.etdm import ETDM
from gpaw.solvation import SolvationGPAW, get_HW14_water_kwargs


def test_solvation_water_water_etdm_lcao():
    SKIP_VAC_CALC = True

    h = 0.24
    vac = 4.0
    convergence = {
        'energy': 0.05 / 8.,
        'density': 10.,
        'eigenstates': 10.,
    }

    atoms = Cluster(molecule('H2O'))
    atoms.minimal_box(vac, h)

    if not SKIP_VAC_CALC:
        atoms.calc = GPAW(mode='lcao', xc='PBE', h=h, basis='dzp',
                          occupations={'name': 'fixed-uniform'},
                          eigensolver='etdm',
                          mixer={'backend': 'no-mixing'},
                          nbands='nao', symmetry='off',
                          convergence=convergence)
        Evac = atoms.get_potential_energy()
        print(Evac)
    else:
        # h=0.24, vac=4.0, setups: 0.9.20000, convergence: only energy 0.05 / 8
        Evac = -12.68228003345474

    atoms.calc = SolvationGPAW(mode='lcao', xc='PBE', h=h, basis='dzp',
                               occupations={'name': 'fixed-uniform'},
                               eigensolver=ETDM(
                                   linesearch_algo={'name': 'max-step'}),
                               mixer={'backend': 'no-mixing'},
                               nbands='nao', symmetry='off',
                               convergence=convergence,
                               **get_HW14_water_kwargs())
    Ewater = atoms.get_potential_energy()
    Eelwater = atoms.calc.get_electrostatic_energy()
    Esurfwater = atoms.calc.get_solvation_interaction_energy('surf')
    DGSol = (Ewater - Evac) / (kcal / mol)
    print('Delta Gsol: %s kcal / mol' % DGSol)

    equal(DGSol, -6.3, 2.)
    equal(Ewater, Eelwater + Esurfwater, 1e-14)



"""Make sure we get an exception when an atom is too close to the boundary."""
from ase import Atoms
from gpaw import GPAW
from gpaw.grid_descriptor import GridBoundsError
from gpaw.utilities import AtomsTooClose
import pytest


@pytest.mark.parametrize('mode', ['fd', 'pw'])
def test_too_close_to_boundary(mode, gpaw_new):
    if mode == 'pw' and not gpaw_new:
        return
    a = 4.0
    x = 0.1
    hydrogen = Atoms('H', [(x, x, x)],
                     cell=(a, a, a),
                     pbc=(1, 1, 0))
    hydrogen.calc = GPAW(mode=mode)
    with pytest.raises((GridBoundsError, AtomsTooClose)):
        hydrogen.get_potential_energy()



"""Test calculation for bare proton.

Also, its interaction with an external potential in the form of a point charge
is tested.
"""
import numpy as np
import pytest
from ase import Atoms
from ase.units import Bohr, Hartree

from gpaw import GPAW
from gpaw.external import PointChargePotential


@pytest.mark.later
def test_generic_proton(in_tmp_dir):
    a = 4.5
    H = Atoms('H', [(a / 2, a / 2, a / 2)],
              pbc=0,
              cell=(a, a, a))
    H.calc = GPAW(mode='fd', nbands=1, h=0.2, charge=1, txt='H.txt')
    e0 = H.get_potential_energy()
    assert abs(e0 + H.calc.get_reference_energy()) < 0.014

    # Test the point charge potential with a smooth cutoff:
    pcp = PointChargePotential([-1], rc2=5.5, width=1.5)
    H.calc = H.calc.new(external=pcp)
    E = []
    F = []
    D = np.linspace(2, 6, 30)
    for d in D:
        pcp.set_positions([[a / 2, a / 2, a / 2 + d]])
        e = H.get_potential_energy() - e0
        f1 = H.get_forces()
        f2 = pcp.get_forces(H.calc)
        eref = -1 / d * Bohr * Hartree
        print(d, e, eref, abs(f1 + f2).max())
        if d < 4.0:
            error = e + 1 / d * Bohr * Hartree
            assert abs(error) < 0.01, error
        assert abs(f1 + f2).max() < 0.01
        E.append(e)
        F.append(f1[0, 2])

    E = np.array(E)
    FF = (E[2:] - E[:-2]) / (D[2] - D[0])  # numerical force
    print(abs(F[1:-1] - FF).max())
    assert abs(F[1:-1] - FF).max() < 0.1

    if not True:
        import matplotlib.pyplot as plt
        plt.plot(D, E)
        plt.plot(D, -1 / D * Bohr * Hartree)
        plt.plot(D, F)
        plt.plot(D[1:-1], FF)
        plt.show()



from __future__ import annotations

import argparse
import json
import os
import pickle
import random
import subprocess
import sys
from pathlib import Path
from time import time
from typing import Any, TypeVar, Callable, TYPE_CHECKING

import numpy as np
from ase import Atoms
from ase.build import bulk
from ase.units import Bohr, Ha
from gpaw.calculator import GPAW as OldGPAW
from gpaw.mpi import world
from gpaw.new.ase_interface import GPAW as NewGPAW

if TYPE_CHECKING:
    T = TypeVar('T')
    PickFunc = Callable[[list[T]], list[T]]


def main(args: str | list[str] = None) -> int:
    if isinstance(args, str):
        args = args.split()

    parser = argparse.ArgumentParser()
    parser.add_argument('-r', '--repeat')
    parser.add_argument('-p', '--pbc')
    parser.add_argument('-v', '--vacuum')
    parser.add_argument('-M', '--magmoms')
    parser.add_argument('-k', '--kpts')
    parser.add_argument('-m', '--mode')
    parser.add_argument('-c', '--code')
    parser.add_argument('-n', '--ncores')
    parser.add_argument('-s', '--use-symmetry')
    parser.add_argument('-S', '--spin-polarized')
    parser.add_argument('-x', '--complex')
    parser.add_argument('-i', '--ignore-cache', action='store_true')
    parser.add_argument('-o', '--stdout', action='store_true')
    parser.add_argument('--pickle')
    parser.add_argument('--all', action='store_true')
    parser.add_argument('--fuzz', action='store_true')
    parser.add_argument('system', nargs='*')
    args = parser.parse_intermixed_args(args)

    if args.pickle:
        pckl_file = Path(args.pickle)
        atoms, params, result_file = pickle.loads(pckl_file.read_bytes())
        run2(atoms, params, result_file)
        return 0

    many = args.all or args.fuzz

    if many:
        system_names = args.system or list(systems)
        args.repeat = args.repeat or '1x1x1,2x1x1'
        args.vacuum = args.vacuum or '0.0,4.0'
        args.pbc = args.pbc or '0,1'
        args.mode = args.mode or 'pw,lcao,fd'
        args.code = args.code or 'new,old'
        args.ncores = args.ncores or '1,2,3,4'
        args.kpts = args.kpts or '2.0,3.0'
        args.use_symmetry = args.use_symmetry or '1,0'
        args.complex = args.complex or '0,1'
    else:
        system_names = args.system
        args.repeat = args.repeat or '1x1x1'
        args.vacuum = args.vacuum or '0.0'
        args.pbc = args.pbc or '0'
        args.mode = args.mode or 'pw'
        args.code = args.code or 'new'
        args.ncores = args.ncores or '1'
        args.kpts = args.kpts or '2.0'
        args.use_symmetry = args.use_symmetry or '1'
        args.complex = args.complex or '0'

    if world.size > 1:
        args.ncores = str(world.size)

    repeat_all = [[int(r) for r in rrr.split('x')]
                  for rrr in args.repeat.split(',')]
    vacuum_all = [float(v) for v in args.vacuum.split(',')]
    pbc_all = [bool(int(p)) for p in args.pbc.split(',')]

    magmoms = None if args.magmoms is None else [
        float(m) for m in args.magmoms.split(',')]

    mode_all = args.mode.split(',')
    kpts_all = [[int(k) for k in kpt.split(',')] if ',' in kpt else
                float(kpt)
                for kpt in args.kpts.split(',')]

    code_all = args.code.split(',')
    ncores_all = [int(c) for c in args.ncores.split(',')]
    use_symmetry_all = [bool(int(s)) for s in args.use_symmetry.split(',')]
    complex_all = [bool(int(s)) for s in args.complex.split(',')]

    # spinpol

    if args.fuzz:
        def pick(choises):
            return [random.choice(choises)]
    else:
        def pick(choises):
            return choises

    count = 0
    calculations = {}
    ok = True

    while ok:
        for atoms, atag in create_systems(system_names,
                                          repeat_all,
                                          vacuum_all,
                                          pbc_all,
                                          magmoms,
                                          pick):
            for params, ptag in create_parameters(mode_all,
                                                  kpts_all,
                                                  pick):
                tag = atag + ' ' + ptag

                for extra, xtag in create_extra_parameters(code_all,
                                                           ncores_all,
                                                           use_symmetry_all,
                                                           complex_all,
                                                           pick):
                    params2 = {**params, **extra}
                    result = run(atoms,
                                 params2,
                                 tag + ' ' + xtag,
                                 args.ignore_cache,
                                 args.stdout)
                    ok = check(tag, result, calculations)
                    count += 1
                    if not ok:
                        break
                if not ok:
                    break
            if not ok:
                break

        if not args.fuzz:
            break

    return int(not ok)


def run(atoms: Atoms,
        params: dict[str, Any],
        tag: str,
        ignore_cache: bool = False,
        use_stdout: bool = False) -> dict[str, Any]:
    params = params.copy()
    name, things = tag.split(' ', 1)
    print(f'{name:3} {things}:', end='')
    tag = tag.replace(' ', '')
    folder = Path('fuzz')
    if not folder.is_dir():
        folder.mkdir()
    result_file = folder / f'{tag}.json'
    if not use_stdout:
        params['txt'] = str(result_file.with_suffix('.txt'))
    if not result_file.is_file() or ignore_cache:
        print(' ...', end='', flush=True)
        ncores = params.pop('ncores')
        if ncores == world.size:
            result = run2(atoms, params, result_file)
        else:
            pckl_file = result_file.with_suffix('.pckl')
            pckl_file.write_bytes(pickle.dumps((atoms, params, result_file)))
            args = ['mpiexec', '-np', str(ncores),
                    sys.executable, '-m', 'gpaw.test.fuzz',
                    '--pickle', str(pckl_file)]
            extra = os.environ.get('GPAW_MPI_OPTIONS')
            if extra:
                args[1:1] = extra.split()
            subprocess.run(args, check=True, env=os.environ)
            result, _ = json.loads(result_file.read_text())
            pckl_file.unlink()
    else:
        print('    ', end='')
        result, _ = json.loads(result_file.read_text())
    print(f' {result["energy"]:14.6f} eV, {result["time"]:9.3f} s')
    return result


def run2(atoms: Atoms,
         params: dict[str, Any],
         result_file: Path) -> dict[str, Any]:
    params = params.copy()

    code = params.pop('code')
    if code[0] == 'n':
        if params.pop('dtype', None) == complex:
            params['mode']['force_complex_dtype'] = True
        calc = NewGPAW(**params)
    else:
        calc = OldGPAW(**params)
    atoms.calc = calc

    t1 = time()
    energy = atoms.get_potential_energy()
    try:
        forces = atoms.get_forces()
    except NotImplementedError:
        forces = None

    t2 = time()

    result = {'time': t2 - t1,
              'energy': energy,
              'forces': None if forces is None else forces.tolist()}

    gpw_file = result_file.with_suffix('.gpw')
    calc.write(gpw_file, mode='all')

    calculation = NewGPAW(gpw_file).calculation

    energy2 = calculation.results['energy'] * Ha
    assert abs(energy2 - energy) < 1e-13, (energy2, energy)

    if forces is not None:
        forces2 = calculation.results['forces'] * Ha / Bohr
        assert abs(forces2 - forces).max() < 1e-14

    # ibz_index = atoms.calc.wfs.kd.bz2ibz_k[p.kpt]
    # eigs = atoms.calc.get_eigenvalues(ibz_index, p.spin)

    if world.rank == 0:
        if 'dtype' in params:
            params['dtype'] = 'complex'
        result_file.write_text(json.dumps([result, params], indent=2))

    atoms.calc = None

    return result


def check(tag: str,
          result: dict[str, Any],
          calculations: dict[str, dict[str, Any]]) -> bool:
    if tag not in calculations:
        calculations[tag] = result
        return True

    result0 = calculations[tag]
    e0 = result0['energy']
    f0 = result0['forces']
    e = result['energy']
    f = result['forces']
    error = e - e0
    if abs(error) > 0.0005:
        print('Energy error:', e, e0, error)
        return False
    if f0 is None:
        if f is not None:
            calculations[tag]['forces'] = f
        return True
    if f is not None:
        error = abs(np.array(f) - f0).max()
        if error > 0.001:
            print('Force error:', error)
            return False
    return True


def create_systems(system_names: list[str],
                   repeats: list[list[int]],
                   vacuums: list[float],
                   pbcs: list[bool],
                   magmoms: list[float] | None,
                   pick: PickFunc) -> tuple[Atoms, str]:
    for name in pick(system_names):
        atoms = systems[name]
        for repeat in pick(repeats):
            if any(not p and r > 1 for p, r in zip(atoms.pbc, repeat)):
                continue
            ratoms = atoms.repeat(repeat)
            for vacuum in pick(vacuums):
                if vacuum:
                    vatoms = ratoms.copy()
                    axes = [a for a, p in enumerate(atoms.pbc) if not p]
                    if axes:
                        vatoms.center(vacuum=vacuum, axis=axes)
                    else:
                        continue
                else:
                    vatoms = ratoms
                for pbc in pick(pbcs):
                    if pbc:
                        if vatoms.pbc.all():
                            continue
                        patoms = vatoms.copy()
                        patoms.pbc = pbc
                    else:
                        patoms = vatoms

                    if magmoms is not None:
                        patoms.set_initial_magnetic_moments(
                            magmoms * (len(patoms) // len(magmoms)))

                    tag = (f'{name} '
                           f'-r{"x".join(str(r) for r in repeat)} '
                           f'-v{vacuum:.1f} '
                           f'-p{int(pbc)}')
                    yield patoms, tag


def create_parameters(modes: list[str],
                      kpts_all: list[float | list[int]],
                      pick: PickFunc) -> tuple[dict[str, Any], str]:
    for mode in pick(modes):
        for kpt in pick(kpts_all):
            if isinstance(kpt, float):
                kpts = {'density': kpt}
                ktag = f'-k{kpt:.1f}'
            else:
                kpts = kpt
                ktag = f'-k{"x".join(str(k) for k in kpt)}'
            yield {'mode': mode,
                   'kpts': kpts}, f'-m{mode} {ktag}'


def create_extra_parameters(codes: list[str],
                            ncores_all: list[int],
                            symmetry_all: list[bool],
                            complex_all: list[bool],
                            pick: PickFunc) -> dict[str, Any]:
    for code in pick(codes):
        for ncores in pick(ncores_all):
            params = {'code': code,
                      'ncores': ncores}
            for use_symm in pick(symmetry_all):
                if not use_symm:
                    sparams = {**params, 'symmetry': 'off'}
                else:
                    sparams = params
                for force_complex_dtype in pick(complex_all):
                    if force_complex_dtype:
                        sparams['dtype'] = complex
                    yield (sparams,
                           (f'-c{code} -n{ncores} -s{int(use_symm)} '
                            f'-x{int(force_complex_dtype)}'))


systems = {}


def system(func):
    systems[func.__name__] = func()
    return func


@system
def h():
    atoms = Atoms('H', magmoms=[1])
    atoms.center(vacuum=2.0)
    return atoms


@system
def h2():
    atoms = Atoms('H2', [(0, 0, 0), (0, 0.75, 0)])
    atoms.center(vacuum=2.0)
    return atoms


@system
def si():
    atoms = bulk('Si', a=5.4)
    return atoms


@system
def fe():
    atoms = bulk('Fe')
    atoms.set_initial_magnetic_moments([2.3])
    return atoms


@system
def li():
    L = 5.0
    atoms = Atoms('Li', cell=[L, L, 1.5], pbc=(0, 0, 1))
    atoms.center()
    return atoms


if __name__ == '__main__':
    raise SystemExit(main())



import numpy as np
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.grid_descriptor import GridDescriptor
from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
from gpaw.spline import Spline
from gpaw.setup import Setup


def test_fd_ops_nabla():
    rc = 2.0
    a = 2.5 * rc
    n = 64
    lmax = 2
    b = 8.0
    m = (lmax + 1)**2
    gd = GridDescriptor([n, n, n], [a, a, a])
    r = np.linspace(0, rc, 200)
    g = np.exp(-(r / rc * b)**2)
    splines = [Spline(l=l, rmax=rc, f_g=g) for l in range(lmax + 1)]
    c = LFC(gd, [splines])
    c.set_positions([(0, 0, 0)])
    psi = gd.zeros(m)
    d0 = c.dict(m)
    if 0 in d0:
        d0[0] = np.identity(m)
    c.add(psi, d0)
    d1 = c.dict(m, derivative=True)
    c.derivative(psi, d1)

    class TestSetup(Setup):
        l_j = range(lmax + 1)
        nj = lmax + 1
        ni = m

        def __init__(self):
            pass
    rgd = EquidistantRadialGridDescriptor(r[1], len(r))
    g = [np.exp(-(r / rc * b)**2) * r**l for l in range(lmax + 1)]
    d2 = TestSetup().get_derivative_integrals(rgd, g, np.zeros_like(g))
    if 0 in d1:
        print(abs(d1[0] - d2).max())
        assert abs(d1[0] - d2).max() < 2e-9




import numpy as np
from scipy.linalg import eigh

from gpaw import debug


class MatrixDescriptor:
    """Class representing a 2D matrix shape.  Base class for parallel
    matrix descriptor with BLACS."""

    def __init__(self, M, N):
        self.shape = (M, N)

    def __bool__(self):
        return self.shape[0] != 0 and self.shape[1] != 0

    def zeros(self, n=(), dtype=float):
        """Return array of zeroes with the correct size on all CPUs.

        The last two dimensions will be equal to the shape of this
        descriptor.  If specified as a tuple, can have any preceding
        dimension."""
        return self._new_array(np.zeros, n, dtype)

    def empty(self, n=(), dtype=float):
        """Return array of zeros with the correct size on all CPUs.

        See zeros()."""
        return self._new_array(np.empty, n, dtype)

    def _new_array(self, func, n, dtype):
        if isinstance(n, int):
            n = n,
        shape = n + self.shape
        return func(shape, dtype)

    def check(self, a_mn):
        """Check that specified array is compatible with this descriptor."""
        return a_mn.shape == self.shape and a_mn.flags.contiguous

    def checkassert(self, a_mn):
        ok = self.check(a_mn)
        if not ok:
            if not a_mn.flags.contiguous:
                msg = 'Matrix with shape %s is not contiguous' % (a_mn.shape,)
            else:
                msg = ('%s-descriptor incompatible with %s-matrix' %
                       (self.shape, a_mn.shape))
            raise AssertionError(msg)

    def general_diagonalize_dc(self, H_mm, S_mm, C_mm, eps_M,
                               UL='L', iu=None):
        if iu is None:
            n = len(H_mm)
            eigvals = None  # all
        else:
            n = iu
            eigvals = (0, n - 1)
        eps_M[:n], C_mm.T[:, :n] = eigh(H_mm, S_mm,
                                        subset_by_index=eigvals,
                                        overwrite_a=True,
                                        check_finite=debug)
        if C_mm.dtype == complex:
            np.negative(C_mm.imag, C_mm.imag)

    def my_blocks(self, array_mn):
        yield (0, self.shape[0], 0, self.shape[1], array_mn)

    def estimate_memory(self, mem, dtype):
        """Handled by subclass."""
        pass


class BandMatrixDescriptor(MatrixDescriptor):
    """Descriptor-class for square matrices of bands times bands."""

    def __init__(self, bd, gd, ksl):
        MatrixDescriptor.__init__(self, bd.nbands, bd.nbands)
        self.bd = bd
        self.gd = gd  # XXX used?
        self.ksl = ksl  # not really used...

    def assemble_blocks(self, A_qnn, A_NN, hermitian):
        """Assign all distributed sub-blocks pertaining from various rank to
        the relevant parts of a Hermitian or non-Hermitian matrix A_NN.

        Parameters:

        A_qnn: ndarray
            Sub-blocks belonging to the specified rank.
        A_NN: ndarray
            Full matrix in which to write contributions from sub-blocks.
        hermitian: bool
            Indicates whether A_NN is a Hermitian matrix, in which
            case only the lower triangular part is assigned to.

        Note that the sub-block buffers are used for communicating across the
        band communicator, hence A_qnn will be altered during the assembly.
        """
        if self.bd.comm.size == 1:
            if hermitian:
                self.triangular_blockwise_assign(A_qnn, A_NN, 0)
            else:
                self.full_blockwise_assign(A_qnn, A_NN, 0)
            return

        if self.bd.comm.rank == 0:
            for band_rank in range(self.bd.comm.size):
                if band_rank > 0:
                    self.bd.comm.receive(A_qnn, band_rank, 13)
                if hermitian:
                    self.triangular_blockwise_assign(A_qnn, A_NN, band_rank)
                else:
                    self.full_blockwise_assign(A_qnn, A_NN, band_rank)
        else:
            self.bd.comm.send(A_qnn, 0, 13)

    def triangular_blockwise_assign(self, A_qnn, A_NN, band_rank):
        """Assign the sub-blocks pertaining from a given rank to the lower
        triangular part of a Hermitian matrix A_NN. This subroutine is used
        for matrix assembly.

        Parameters:

        A_qnn: ndarray
            Sub-blocks belonging to the specified rank.
        A_NN: ndarray
            Full matrix in which to write contributions from sub-blocks.
        band_rank: int
            Communicator rank to which the sub-blocks belongs.

        Note that a Hermitian matrix requires Q=B//2+1 blocks of M x M
        elements where B is the communicator size and M=N//B for N bands.
        """
        N = self.bd.mynbands
        B = self.bd.comm.size
        assert band_rank in range(B)

        if B == 1:
            # Only fill in the lower part
            mask = np.tri(N).astype(bool)
            A_NN[mask] = A_qnn.reshape((N, N))[mask]
            return

        # A_qnn[q2,myn1,myn2] on rank q1 is the q2'th overlap calculated
        # between <psi_n1| and A|psit_n2> where n1 <-> (q1,myn1) and
        # n2 <-> ((q1+q2)%B,myn2) since we've sent/received q2 times.
        q1 = band_rank
        Q = B // 2 + 1
        if debug:
            assert A_qnn.shape == (Q, N, N)

        # Note that for integer inequalities, these relations are useful (X>0):
        #     A*X > B   <=>   A > B//X   ^   A*X <= B   <=>   A <= B//X

        if self.bd.strided:
            A_nbnb = A_NN.reshape((N, B, N, B))
            mask = np.empty((N, N), dtype=bool)
            for q2 in range(Q):
                # n1 = (q1+q2)%B + myn1*B   ^   n2 = q1 + myn2*B
                #
                # We seek the lower triangular part i.e. n1 >= n2
                #   <=>   (myn2-myn1)*B <= (q1+q2)%B-q1
                #   <=>   myn2-myn1 <= dq//B
                dq = (q1 + q2) % B - q1  # within ]-B; Q[ so dq//B is -1 or 0

                # Create mask for lower part of current block
                mask[:] = np.tri(N, N, dq // B)
                if debug:
                    m1, m2 = np.indices((N, N))
                    assert dq in range(-B + 1, Q)
                    assert (mask == (m1 >= m2 - dq // B)).all()

                # Copy lower part of A_qnn[q2] to its rightfull place
                A_nbnb[:, (q1 + q2) % B, :, q1][mask] = A_qnn[q2][mask]

                # Negate the transposed mask to get complementary mask
                mask = ~mask.T

                # Copy upper part of Hermitian conjugate of A_qnn[q2]
                A_nbnb[:, q1, :,
                       (q1 + q2) % B][mask] = A_qnn[q2].T.conj()[mask]
        else:
            A_bnbn = A_NN.reshape((B, N, B, N))

            # Optimization for the first block
            if q1 == 0:
                A_bnbn[:Q, :, 0] = A_qnn
                return

            for q2 in range(Q):
                # n1 = ((q1+q2)%B)*N + myn1   ^   n2 = q1*N + myn2
                #
                # We seek the lower triangular part i.e. n1 >= n2
                #   <=>   ((q1+q2)%B-q1)*N >= myn2-myn1
                #   <=>   myn2-myn1 <= dq*N
                #   <=>   entire block if dq > 0,
                #   ...   myn2 <= myn1 if dq == 0,
                #   ...   copy nothing if dq < 0
                if q1 + q2 < B:
                    A_bnbn[q1 + q2, :, q1] = A_qnn[q2]
                else:
                    A_bnbn[q1, :, q1 + q2 - B] = A_qnn[q2].T.conj()

    def full_blockwise_assign(self, A_qnn, A_NN, band_rank):
        """Assign the sub-blocks pertaining from a given rank to the full
        non-Hermitian matrix A_NN. This subroutine is used for matrix assembly.

        Parameters:

        A_qnn: ndarray
            Sub-blocks belonging to the specified rank.
        A_NN: ndarray
            Full matrix, in which to write contributions from sub-blocks.
        band_rank: int
            Communicator rank to which the sub-blocks belongs.

        Note that a non-Hermitian matrix requires Q=B blocks of M x M
        elements where B is the communicator size and M=N//B for N bands.
        """
        N = self.bd.mynbands
        B = self.bd.comm.size
        assert band_rank in range(B)

        if B == 1:
            A_NN[:] = A_qnn.reshape((N, N))
            return

        # A_qnn[q2,myn1,myn2] on rank q1 is the q2'th overlap calculated
        # between <psi_n1| and A|psit_n2> where n1 <-> (q1,myn1) and
        # n2 <-> ((q1+q2)%B,myn2) since we've sent/received q2 times.
        q1 = band_rank
        Q = B
        if debug:
            assert A_qnn.shape == (Q, N, N)

        if self.bd.strided:
            A_nbnb = A_NN.reshape((N, B, N, B))
            for q2 in range(Q):
                A_nbnb[:, (q1 + q2) % B, :, q1] = A_qnn[q2]
        else:
            A_bnbn = A_NN.reshape((B, N, B, N))

            # Optimization for the first block
            if q1 == 0:
                A_bnbn[:Q, :, 0] = A_qnn
                return

            for q2 in range(Q):
                A_bnbn[(q1 + q2) % B, :, q1] = A_qnn[q2]

    def extract_block(self, A_NN, q1, q2):
        """Extract the sub-block pertaining from a given pair of ranks within
        the full matrix A_NN. Extraction may result in copies to assure unit
        stride, thus one should not utilize this routine for altering A_NN.

        Parameters:

        A_NN: ndarray
            Full matrix, from which to read the requested sub-block.
        q1: int
            Communicator rank to which the sub-block belongs (row index).
        q2: int
            Communicator rank the sub-block originated from (column index).

        Note that a Hermitian matrix requires just Q=B//2+1 blocks of M x M
        elements where B is the communicator size and M=N//B for N bands.
        Therefor, care should be taken to only request q1,q2 pairs which
        are connected by Q shifts or less if A_NN is lower triangular.
        """
        N = self.bd.mynbands
        B = self.bd.comm.size

        if B == 1:
            return A_NN

        if self.bd.strided:
            A_nbnb = A_NN.reshape((N, B, N, B))
            # last dim must have unit stride
            return A_nbnb[:, q2, :, q1].copy()
        else:
            A_bnbn = A_NN.reshape((B, N, B, N))
            return A_bnbn[q2, :, q1]

    def redistribute_input(self, A_NN):  # do nothing
        return A_NN

    def redistribute_output(self, A_NN):  # do nothing
        if debug:
            self.checkassert(A_NN)
        return A_NN

    def estimate_memory(self, mem, dtype):
        # Temporary work arrays included in estimate #
        nbands = self.bd.nbands
        itemsize = mem.itemsize[dtype]
        mem.subnode('A_NN', nbands * nbands * itemsize)


class BlacsBandMatrixDescriptor(MatrixDescriptor):
    """Descriptor-class for square BLACS matrices of bands times bands."""

    def __init__(self, bd, gd, ksl):
        # XXX a hack...
        MatrixDescriptor.__init__(self, bd.nbands, bd.mynbands)
        self.bd = bd
        self.gd = gd  # XXX used?
        self.ksl = ksl

    def assemble_blocks(self, A_qnn, A_Nn, hermitian):
        """Assign all distributed sub-blocks pertaining from various rank to
        the relevant parts of a Hermitian or non-Hermitian matrix A_NN.

        Parameters:

        A_qnn: ndarray
            Sub-blocks belonging to the specified rank.
        A_Nn: ndarray
            Full column vector in which to write contributions from sub-blocks.
        hermitian: bool
            Indicates whether A_Nn represents a Hermitian matrix, in which
            case only the lower triangular part is assigned to.

        Note that the sub-block buffers are used for communicating across the
        band communicator, hence A_qnn will be altered during the assembly.
        """

        band_rank = self.bd.comm.rank
        if hermitian:
            self.triangular_columnwise_assign(A_qnn, A_Nn, band_rank)
        else:
            self.full_columnwise_assign(A_qnn, A_Nn, band_rank)

    def triangular_columnwise_assign(self, A_qnn, A_Nn, band_rank):
        """Assign the sub-blocks pertaining from a given rank to the lower
        triangular part of a Hermitian matrix A_NN. This subroutine is used
        for matrix assembly.

        Parameters:

        A_qnn: ndarray
            Sub-blocks belonging to the specified rank.
        A_Nn: ndarray
            Full column vector in which to write contributions from sub-blocks.
        band_rank: int
            Communicator rank to which the sub-blocks belongs.

        Note that a Hermitian matrix requires Q=B//2+1 blocks of M x M
        elements where B is the communicator size and M=N//B for N bands.
        """
        N = self.bd.mynbands
        B = self.bd.comm.size
        assert band_rank in range(B)

        if B == 1:
            # Only fill in the lower part
            mask = np.tri(N).astype(bool)
            A_Nn[mask] = A_qnn.reshape((N, N))[mask]
            return

        # A_qnn[q2,myn1,myn2] on rank q1 is the q2'th overlap calculated
        # between <psi_n1| and A|psit_n2> where n1 <-> (q1,myn1) and
        # n2 <-> ((q1+q2)%B,myn2) since we've sent/received q2 times.
        q1 = band_rank
        Q = B // 2 + 1
        if debug:
            assert A_qnn.shape == (Q, N, N)

        # Note that for integer inequalities, these relations are useful (X>0):
        #     A*X > B   <=>   A > B//X   ^   A*X <= B   <=>   A <= B//X

        if self.bd.strided:
            raise NotImplementedError

            """
            A_nbn = A_NN.reshape((N, B, N))
            mask = np.empty((N,N), dtype=bool)
            for q2 in range(Q):
                # n1 = (q1+q2)%B + myn1*B   ^   n2 = q1 + myn2*B
                #
                # We seek the lower triangular part i.e. n1 >= n2
                #   <=>   (myn2-myn1)*B <= (q1+q2)%B-q1
                #   <=>   myn2-myn1 <= dq//B
                dq = (q1+q2)%B-q1 # within ]-B; Q[ so dq//B is -1 or 0

                # Create mask for lower part of current block
                mask[:] = np.tri(N, N, dq//B)
                if debug:
                    m1,m2 = np.indices((N,N))
                    assert dq in range(-B+1,Q)
                    assert (mask == (m1 >= m2 - dq//B)).all()

                # Copy lower part of A_qnn[q2] to its rightfull place
                A_nbn[:, (q1+q2)%B][mask] = A_qnn[q2][mask]

                # Negate the transposed mask to get complementary mask
                mask = ~mask.T

                # Copy upper part of Hermitian conjugate of A_qnn[q2]
                A_nbn[:, q1][mask] = A_qnn[q2].T.conj()[mask]
            """
        else:
            A_bnn = A_Nn.reshape((B, N, N))

            for q2 in range(Q):
                # n1 = ((q1+q2)%B)*N + myn1   ^   n2 = q1*N + myn2
                #
                # We seek the lower triangular part i.e. n1 >= n2
                #   <=>   ((q1+q2)%B-q1)*N >= myn2-myn1
                #   <=>   myn2-myn1 <= dq*N
                #   <=>   entire block if dq > 0,
                #   ...   myn2 <= myn1 if dq == 0,
                #   ...   copy nothing if dq < 0

                if q1 + q2 < B:
                    A_bnn[q1 + q2] = A_qnn[q2]

            reqs = []
            if q1 < Q - 1:  # receive from ranks >= Q
                if debug:
                    Q2 = np.arange(Q, B - q1)
                    print('q1=%d, q2: %12s | recv from q1+q2:%12s -> A_bnn%s' %
                          (q1, Q2.tolist(), (q1 + Q2).tolist(),
                           (q1 + Q2).tolist()))

                for q2 in range(Q, B - q1):
                    rrank = q1 + q2
                    A_nn = A_bnn[q1 + q2]
                    reqs.append(self.bd.comm.receive(A_nn, rrank, block=False))
            elif q1 >= Q:  # send to ranks < Q-1
                if debug:
                    Q2 = np.arange(B - q1, B - Q + 1)[::-1]
                    print('q1=%d, q2: %12s | '
                          'send to q1+q2-B:%12s <- A_qnn%s.T.conj()' %
                          (q1, Q2.tolist(), (q1 + Q2 - B).tolist(),
                           Q2.tolist()))

                # symmetrize comm
                for q2 in reversed(range(B - q1, B - Q + 1)):
                    srank = q1 + q2 - B
                    sbuf_nn = np.ascontiguousarray(np.conjugate(A_qnn[q2].T))
                    reqs.append(self.bd.comm.send(sbuf_nn, srank, block=False))
            else:
                if debug:
                    print('q1=%d, do nothing...' % q1)

            self.bd.comm.waitall(reqs)

    def full_columnwise_assign(self, A_qnn, A_Nn, band_rank):
        """Assign the sub-blocks pertaining from a given rank to the columns of
        non-Hermitian matrix A_Nn. This subroutine is used for column assembly.

        Parameters:

        A_qnn: ndarray
            Sub-blocks belonging to the specified rank.
        A_Nn: ndarray
            Full column vector, in which to write contributions from
            sub-blocks.
        band_rank: int
            Communicator rank to which the sub-blocks belongs.

        Note that a non-Hermitian matrix requires Q=B blocks of M x M
        elements where B is the communicator size and M=N//B for N bands.
        """
        N = self.bd.mynbands
        B = self.bd.comm.size
        assert band_rank in range(B)

        if B == 1:
            A_Nn[:] = A_qnn.reshape((N, N))
            return

        # A_qnn[q2,myn1,myn2] on rank q1 is the q2'th overlap calculated
        # between <psi_n1| and A|psit_n2> where n1 <-> (q1,myn1) and
        # n2 <-> ((q1+q2)%B,myn2) since we've sent/received q2 times.
        q1 = band_rank
        Q = B

        if self.bd.strided:
            A_nbn = A_Nn.reshape((N, B, N))
            for q2 in range(Q):
                A_nbn[:, (q1 + q2) % B] = A_qnn[q2]
        else:
            A_bnn = A_Nn.reshape((B, N, N))

            # Optimization for the first block
            if q1 == 0:
                A_bnn[:Q] = A_qnn
                return

            for q2 in range(Q):
                A_bnn[(q1 + q2) % B] = A_qnn[q2]

    def extract_block_from_column(self, A_Nn, q1, q2):
        """Extract the sub-block pertaining from a given pair of ranks within
        the full matrix A_NN. Extraction may result in copies to assure unit
        stride, thus one should not utilize this routine for altering A_NN.

        Parameters:

        A_Nn: ndarray
            Full column vector, from which to read the requested sub-block.
        q1: int
            Communicator rank to which the sub-block belongs (row index).
        q2: int
            Communicator rank the sub-block originated from (column index).

        Note that a Hermitian matrix requires just Q=B//2+1 blocks of M x M
        elements where B is the communicator size and M=N//B for N bands.
        Therefor, care should be taken to only request q1,q2 pairs which
        are connected by Q shifts or less if A_NN is lower triangular.
        """
        N = self.bd.mynbands
        B = self.bd.comm.size

        if B == 1:
            return A_Nn

        if q1 == self.bd.comm.rank:  # XXX must evaluate the same on all ranks!
            if self.bd.strided:
                A_nbn = A_Nn.reshape((N, B, N))
                return A_nbn[:, q2, :].copy()  # block must be contiguous
            else:
                A_bnn = A_Nn.reshape((B, N, N))
                return A_bnn[q2]

        # Time for us to put the cards on the table
        Qs = np.empty((self.bd.comm.size, 2), dtype=int)
        self.bd.comm.all_gather(np.array([q1, q2]), Qs)
        Q1, Q2 = Qs.T

        # Block exchange. What do we want and who should we be sending to?
        rrank = q1
        srank = np.argwhere(Q1 == self.bd.comm.rank).ravel().item()
        sq2 = Q2[srank]

        if debug:
            S = np.empty(self.bd.comm.size, dtype=int)
            self.bd.comm.all_gather(np.array([srank]), S)
            if self.bd.comm.rank == 0:
                print('recv(Q1): %s, send(Q1==rank): %s' %
                      (Q1.tolist(), S.tolist()))

        if self.bd.strided:
            A_nbn = A_Nn.reshape((N, B, N))
            sbuf_nn = A_nbn[:, sq2, :].copy()  # block must be contiguous
        else:
            A_bnn = A_Nn.reshape((B, N, N))
            sbuf_nn = A_bnn[sq2]

        A_nn = np.empty_like(sbuf_nn)
        self.bd.comm.sendreceive(sbuf_nn, srank, A_nn, rrank)
        return A_nn

    def extract_block_from_row(self, A_nN, q1, q2):
        """Extract the sub-block pertaining from a given pair of ranks within
        the full matrix A_NN. Extraction may result in copies to assure unit
        stride, thus one should not utilize this routine for altering A_NN.

        Parameters:

        A_nN: ndarray
            Full row vector, from which to read the requested sub-block.
        q1: int
            Communicator rank to which the sub-block belongs (row index).
        q2: int
            Communicator rank the sub-block originated from (column index).

        Note that a Hermitian matrix requires just Q=B//2+1 blocks of M x M
        elements where B is the communicator size and M=N//B for N bands.
        Therefor, care should be taken to only request q1,q2 pairs which
        are connected by Q shifts or less if A_NN is lower triangular.
        """
        N = self.bd.mynbands
        B = self.bd.comm.size

        if B == 1:
            return A_nN

        if q2 == self.bd.comm.rank:  # XXX must evaluate the same on all ranks!
            if self.bd.strided:
                # XXX This case seems to be untested
                A_nnb = A_nN.reshape((N, N, B))
                return A_nnb[..., q1].copy('C')  # block must be contiguous
            else:
                A_nbn = A_nN.reshape((N, B, N))
                return A_nbn[:, q1, :].copy('C')  # block must be contiguous
        else:
            raise NotImplementedError

    extract_block = extract_block_from_row  # XXX ugly but works

    def redistribute_input(self, A_nn, A_nN=None):  # 2D -> 1D row layout
        if A_nN is None:
            A_nN = self.ksl.nNdescriptor.empty(dtype=A_nn.dtype)
        self.ksl.nn2nN.redistribute(A_nn, A_nN)
        if not self.ksl.nNdescriptor.blacsgrid.is_active():
            assert A_nN.shape == (0, 0)
            A_nN = np.empty((self.bd.mynbands, self.bd.nbands),
                            dtype=A_nN.dtype)
        self.gd.comm.broadcast(A_nN, 0)
        return A_nN

    def redistribute_output(self, A_Nn, A_nn=None):  # 1D column -> 2D layout
        if not self.ksl.Nndescriptor.blacsgrid.is_active():
            A_Nn = np.empty((0, 0), dtype=A_Nn.dtype)
        if A_nn is None:
            A_nn = self.ksl.nndescriptor.empty(dtype=A_Nn.dtype)
        self.ksl.Nn2nn.redistribute(A_Nn, A_nn)
        return A_nn

    def estimate_memory(self, mem, dtype):
        # Temporary work arrays included in estimate #
        mynbands = self.bd.mynbands
        nbands = self.bd.nbands
        itemsize = mem.itemsize[dtype]
        mem.subnode('2 A_nN', 2 * mynbands * nbands * itemsize)
        mem.subnode('2 A_nn', 2 * nbands * nbands / self.ksl.blockgrid.ncpus *
                    itemsize)



from __future__ import annotations

from math import pi
from typing import TYPE_CHECKING

import numpy as np
from ase.units import Bohr

from gpaw.core.atom_arrays import AtomArrays, AtomArraysLayout
from gpaw.core.uniform_grid import UGArray
from gpaw.setup import Setups
from gpaw.spherical_harmonics import Y
from gpaw.spline import Spline
from gpaw.typing import Array1D, Array3D, Vector, Array2D
from gpaw.new import zips

if TYPE_CHECKING:
    from gpaw.new.calculation import DFTCalculation


class Densities:
    def __init__(self,
                 nt_sR: UGArray,
                 D_asii: AtomArrays,
                 fracpos_ac: Array2D,
                 setups: Setups):
        self.nt_sR = nt_sR
        self.D_asii = D_asii
        self.fracpos_ac = fracpos_ac
        self.setups = setups

    @classmethod
    def from_calculation(cls, calculation: DFTCalculation):
        density = calculation.state.density
        return cls(density.nt_sR,
                   density.D_asii,
                   calculation.fracpos_ac,
                   calculation.setups)

    def pseudo_densities(self,
                         grid_spacing: float = None,  # Ang
                         grid_refinement: int = None,
                         add_compensation_charges: bool = True
                         ) -> UGArray:
        nt_sR = self._pseudo_densities(grid_spacing, grid_refinement)

        ncomponents = nt_sR.dims[0]
        ndensities = ncomponents % 3

        if add_compensation_charges:
            cc_asL = AtomArraysLayout(
                [(ncomponents, setup.Delta_iiL.shape[2])
                 for setup in self.setups],
                atomdist=self.D_asii.layout.atomdist).empty()

            for a, D_sii in self.D_asii.items():
                Q_sL = np.einsum('sij, ijL -> sL',
                                 D_sii, self.setups[a].Delta_iiL)
                delta = (self.setups[a].Delta0 +
                         self.setups[a].Nv / (4 * pi)**0.5)
                Q_sL[:ndensities, 0] += delta / ndensities
                cc_asL[a] = Q_sL

            ghat_aLR = self.setups.create_compensation_charges(
                nt_sR.desc,
                self.fracpos_ac,
                self.D_asii.layout.atomdist)
            ghat_aLR.add_to(nt_sR, cc_asL)

        return nt_sR.scaled(Bohr, Bohr**-3)

    def _pseudo_densities(self,
                          grid_spacing: float = None,  # Ang
                          grid_refinement: int = None,
                          ) -> UGArray:
        nt_sR = self.nt_sR.to_pbc_grid()
        grid = nt_sR.desc
        if grid_spacing is not None:
            assert grid_refinement is None
            grid = grid.uniform_grid_with_grid_spacing(
                grid_spacing / Bohr)
        elif grid_refinement is not None and grid_refinement > 1:
            grid = grid.new(size=grid.size * grid_refinement)
        else:
            return nt_sR.copy()

        return nt_sR.interpolate(grid=grid)

    def all_electron_densities(self,
                               *,
                               grid_spacing: float = None,  # Ang
                               grid_refinement: int = None,
                               skip_core: bool = False,
                               ) -> UGArray:
        n_sR = self._pseudo_densities(grid_spacing, grid_refinement)
        ncomponents = n_sR.dims[0]
        nspins = ncomponents % 3
        grid = n_sR.desc

        splines = {}
        for a, D_sii in self.D_asii.items():
            fracpos_c = self.fracpos_ac[a]
            setup = self.setups[a]
            if setup not in splines:
                phi_j, phit_j, nc, nct = setup.get_partial_waves()[:4]
                if skip_core:
                    nc = Spline(0, 10.0, [0.0, 0.0])
                rcut = max(setup.rcut_j)
                splines[setup] = (rcut, phi_j, phit_j, nc, nct)
            rcut, phi_j, phit_j, nc, nct = splines[setup]

            # Expected integral of PAW correction:
            electrons_s = np.zeros(ncomponents)
            if skip_core:
                electrons_s[:nspins] = -setup.Nct / nspins
            else:
                electrons_s[:nspins] = (setup.Nc - setup.Nct) / nspins
            electrons_s += (4 * pi)**0.5 * np.einsum('sij, ij -> s',
                                                     D_sii,
                                                     setup.Delta_iiL[:, :, 0])

            # Add PAW correction:
            R_v = fracpos_c @ grid.cell_cv
            electrons_s -= add(R_v, n_sR, phi_j, phit_j, nc, nct, rcut, D_sii)

            if not skip_core:
                # Add missing charge to grid point closest to atom:
                R_c = np.around(grid.size * fracpos_c).astype(int) % grid.size
                R_c -= grid.start_c
                if (R_c >= 0).all() and (R_c < grid.mysize_c).all():
                    for n_R, e in zips(n_sR.data, electrons_s):
                        n_R[tuple(R_c)] += e / grid.dv

        return n_sR.scaled(Bohr, Bohr**-3)


def add(R_v: Vector,
        a_sR: UGArray,
        phi_j: list[Spline],
        phit_j: list[Spline],
        nc: Spline,
        nct: Spline,
        rcut: float,
        D_sii: Array3D) -> Array1D:
    """Add PAW corrections to real-space grid.

    Returns number of elctrons added.
    """
    ug = a_sR.desc
    R_Rv = ug.xyz()
    lmax = max(phi.l for phi in phi_j)
    ncomponents = a_sR.dims[0]
    nspins = ncomponents % 3
    electrons_s = np.zeros(ncomponents)
    start_c = 0 - ug.pbc
    stop_c = 1 + ug.pbc
    for u0 in range(start_c[0], stop_c[0]):
        for u1 in range(start_c[1], stop_c[1]):
            for u2 in range(start_c[2], stop_c[2]):
                d_Rv = R_Rv - (R_v + (u0, u1, u2) @ ug.cell_cv)
                d_R = (d_Rv**2).sum(3)**0.5
                mask_R = d_R < rcut
                npoints = mask_R.sum()
                if npoints == 0:
                    continue

                a_sr = np.zeros((ncomponents, npoints))
                d_rv = d_Rv[mask_R]
                d_r = d_R[mask_R]
                Y_Lr = [Y(L, *d_rv.T) for L in range((lmax + 1)**2)]
                phi_jr = [phi.map(d_r) for phi in phi_j]
                phit_jr = [phit.map(d_r) for phit in phit_j]
                l_j = [phi.l for phi in phi_j]

                i1 = 0
                for l1, phi1_r, phit1_r in zips(l_j, phi_jr, phit_jr):
                    i2 = 0
                    i1b = i1 + 2 * l1 + 1
                    D_smi = D_sii[:, i1:i1b]
                    for l2, phi2_r, phit2_r in zips(l_j, phi_jr, phit_jr):
                        i2b = i2 + 2 * l2 + 1
                        D_smm = D_smi[:, :, i2:i2b]
                        b_sr = np.einsum(
                            'smn, mr, nr -> sr',
                            D_smm,
                            Y_Lr[l1**2:(l1 + 1)**2],
                            Y_Lr[l2**2:(l2 + 1)**2]) * (
                            phi1_r * phi2_r - phit1_r * phit2_r)
                        a_sr += b_sr
                        i2 = i2b
                    i1 = i1b

                dn_r = nc.map(d_r) - nct.map(d_r)
                a_sr[:nspins] += dn_r * ((4 * pi)**-0.5 / nspins)
                electrons_s += a_sr.sum(1) * a_sR.desc.dv
                a_sR.data[:, mask_R] += a_sr
    return electrons_s



"""Wigner-Seitz truncated coulomb interaction.

See:

    Ravishankar Sundararaman and T. A. Arias:
    Phys. Rev. B 87, 165122 (2013)

    Regularization of the Coulomb singularity in exact exchange by
    Wigner-Seitz truncated interactions: Towards chemical accuracy
    in nontrivial systems
"""
from math import pi

import numpy as np
from scipy.special import erf
from ase.units import Bohr
from ase.utils import seterr

import gpaw.mpi as mpi
from gpaw.fftw import get_efficient_fft_size
from gpaw.grid_descriptor import GridDescriptor


class WignerSeitzTruncatedCoulomb:
    def __init__(self, cell_cv, nk_c):
        self.nk_c = nk_c
        bigcell_cv = cell_cv * nk_c[:, np.newaxis]
        L_c = (np.linalg.inv(bigcell_cv)**2).sum(0)**-0.5
        
        self.rc = 0.5 * L_c.min()
        self.a = 5 / self.rc

        nr_c = [get_efficient_fft_size(2 * int(L * self.a * 3.0))
                for L in L_c]

        self.gd = GridDescriptor(nr_c, bigcell_cv, comm=mpi.serial_comm)
        v_ijk = self.gd.empty()

        pos_ijkv = self.gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))
        corner_xv = np.dot(np.indices((2, 2, 2)).reshape((3, 8)).T, bigcell_cv)

        # Ignore division by zero (in 0,0,0 corner):
        with seterr(invalid='ignore'):
            # Loop over first dimension to avoid too large ndarrays.
            for pos_jkv, v_jk in zip(pos_ijkv, v_ijk):
                # Distances to the 8 corners:
                d_jkxv = pos_jkv[:, :, np.newaxis] - corner_xv
                r_jk = (d_jkxv**2).sum(axis=3).min(2)**0.5
                v_jk[:] = erf(self.a * r_jk) / r_jk

        # Fix 0/0 corner value:
        v_ijk[0, 0, 0] = 2 * self.a / pi**0.5

        self.K_Q = np.fft.fftn(v_ijk) * self.gd.dv
    
    def get_description(self):
        descriptors = []
        descriptors.append('Inner radius for %dx%dx%d Wigner-Seitz cell: '
                           '%.3f Ang' % (tuple(self.nk_c) + (self.rc * Bohr,)))
        descriptors.append('Range-separation parameter: %.3f Ang^-1' % (
            self.a / Bohr))
        descriptors.append('FFT size for calculating truncated Coulomb: '
                           '%dx%dx%d' % tuple(self.gd.N_c))
        
        return '\n'.join(descriptors)
    
    def get_potential(self, pd, q_v=None):
        q_c = pd.kd.bzk_kc[0]
        shift_c = (q_c * self.nk_c).round().astype(int)
        max_c = self.gd.N_c // 2
        K_G = pd.zeros()
        N_c = pd.gd.N_c
        if pd.dtype == complex:
            for G, Q in enumerate(pd.Q_qG[0]):
                Q_c = (np.unravel_index(Q, N_c) + N_c // 2) % N_c - N_c // 2
                Q_c = Q_c * self.nk_c + shift_c
                if (abs(Q_c) < max_c).all():
                    K_G[G] = self.K_Q[tuple(Q_c)]
        else:
            Nr_c = N_c.copy()
            Nr_c[2] = N_c[2] // 2 + 1
            for G, Q in enumerate(pd.Q_qG[0]):
                Q_c = np.array(np.unravel_index(Q, Nr_c))
                Q_c[:2] += N_c[:2] // 2
                Q_c[:2] %= N_c[:2]
                Q_c[:2] -= N_c[:2] // 2
                if (abs(Q_c) < max_c).all():
                    K_G[G] = self.K_Q[tuple(Q_c)]

        qG_Gv = pd.get_reciprocal_vectors(add_q=True)
        if q_v is not None:
            qG_Gv += q_v
        G2_G = np.sum(qG_Gv**2, axis=1)
        # G2_G = pd.G2_qG[0]
        a = self.a
        G0 = G2_G.argmin()
        if G2_G[G0] < 1e-11:
            K0 = K_G[G0] + pi / a**2
        with np.errstate(invalid='ignore'):
            K_G += 4 * pi * (1 - np.exp(-G2_G / (4 * a**2))) / G2_G
        if G2_G[G0] < 1e-11:
            K_G[G0] = K0
        return K_G



from __future__ import annotations

import numpy as np
from ase import Atoms
from gpaw.utilities.gpts import get_number_of_grid_points
from gpaw.wavefunctions.lcao import LCAO
from gpaw.wavefunctions.pw import PW


def obtain_gpts_suggestion(cell_cv, ecut, h, print_suggestions=False):
    """Compare PW and LCAO gpts and returns tighter one.

    When two separate simulations using different calculators need to use the
    same real space grid, like in the electron-phonon coupling case, it might
    be necessary to specify gpts=(...) explicitly. This tool helps with finding
    the correct values for a combination of PW and LCAO mode calculations.

    Parameters
    ----------
    cell_cv: np.ndarray
        cell vector array
    ecut: int, float
        planewave cutoff to be used in PW mode
    h: float
        intended maximal grid spacing in LCAO mode
    print_suggestions: bool
       if True, prints human readable information
    """
    Npw_c = get_number_of_grid_points(cell_cv, mode=PW(ecut))
    Nlcao_c = get_number_of_grid_points(cell_cv, h=h, mode=LCAO())

    Nopt_c = np.maximum(Nlcao_c, (Npw_c / 4 + 0.5).astype(int) * 4)

    if print_suggestions:
        print(f'PW({ecut:3.0f}) -> gpts={list(Npw_c)}')
        print(f'LCAO, h={h:1.3f} -> gpts={list(Nlcao_c)}')
        print(f'Recommended for elph: gpts={list(Nopt_c)}')

        if np.all(Npw_c == Nlcao_c):
            print('  Both sets of gpts the same. No action needed.')
        if np.any(Npw_c < Nopt_c):
            print(f'  Add "gpts={list(Nopt_c)}" to PW mode calculator.')
        if np.any(Nlcao_c < Nopt_c):
            print(f'  Use "gpts={list(Nopt_c)}" instead of "h={h:1.3f}" ' +
                  'in LCAO calculator.')

    return Nopt_c


def main(argv: list[str] = None) -> None:
    import argparse

    from ase.io import read

    parser = argparse.ArgumentParser(
        prog='python3 -m gpaw.utilities.gpts',
        description='Calculate optimal gpts size between PW and LCAO mode.')
    add = parser.add_argument
    add('file', metavar='input-file',
        help='ASE readable structure file.')
    add('-e', '--ecut', type=float,
        help='Cutoff "ecut" used in PW mode.')
    add('-g', '--grid-spacing', type=float,
        help='Maximal grid spacing "h" used in LCAO mode.')
    add('-s', '--super-cell', default='1,1,1',
        help='Supercell size.  Example: "-s 2,2,2".')

    args = parser.parse_intermixed_args(argv)
    if not args.ecut or not args.grid_spacing:
        parser.print_help()
        raise SystemExit

    sc = []
    for value in args.super_cell.split(','):
        sc.append(int(value))

    atoms = read(args.file)
    assert isinstance(atoms, Atoms)
    atoms_sc = atoms * sc
    cell_cv = atoms_sc.get_cell()
    obtain_gpts_suggestion(cell_cv, args.ecut, args.grid_spacing, True)


if __name__ == '__main__':
    main()



import numpy as np
import pytest

from gpaw.fd_operators import Laplace
from gpaw.gpu import cupy as cp
from gpaw.grid_descriptor import GridDescriptor
from gpaw.mpi import world


@pytest.mark.gpu
@pytest.mark.parametrize('pbc', [True, False])
def test_fd_laplace(gpu, pbc):
    if world.size > 4:
        # Grid is so small that domain decomposition cannot exceed 4 domains
        assert world.size % 4 == 0
        group, other = divmod(world.rank, 4)
        ranks = np.arange(4 * group, 4 * (group + 1))
        domain_comm = world.new_communicator(ranks)
    else:
        domain_comm = world

    lat = 8.0
    gd = GridDescriptor((32, 32, 32), (lat, lat, lat),
                        pbc_c=pbc, comm=domain_comm)

    if pbc:
        dtype = complex
        phase = np.ones((3, 2), complex)
    else:
        dtype = float
        phase = None

    # Use Gaussian as input
    x, y, z = gd.get_grid_point_coordinates()
    sigma = 1.5
    mu = lat / 2.0

    a = gd.zeros(dtype=dtype)
    a[:] = np.exp(-((x - mu)**2 + (y - mu)**2 + (z - mu)**2) / (2.0 * sigma))
    # analytic solution
    b_analytic = np.zeros_like(a)
    b_analytic[:] = (((x - mu)**2 + (y - mu)**2 + (z - mu)**2) / sigma**2
                     - 3.0 / sigma) * a

    b = np.zeros_like(a)
    a_gpu = cp.asarray(a)
    b_gpu = cp.zeros_like(a_gpu)

    # Laplace
    Laplace(gd, 1.0, 3, dtype=dtype).apply(a, b, phase_cd=phase)
    Laplace(gd, 1.0, 3, dtype=dtype, xp=cp).apply(a_gpu, b_gpu, phase_cd=phase)
    b_ref = b_gpu.get()

    assert b == pytest.approx(b_ref, abs=1e-12)
    # Neglect boundaries in check to analytic solution
    assert (b_analytic[2:-2, 2:-2, 2:-2]
            == pytest.approx(b_ref[2:-2, 2:-2, 2:-2], abs=1e-2))



"""Appelbaum-Hamann local pseudo-potential for silicon.

See::

  Self-Consistent Pseudopotential for Si
  Joel A. Appelbaum and D. R. Hamann
  PRB 8, 1777 (1973)

"""

import numpy as np

from gpaw.setup import BaseSetup
from gpaw.spline import Spline
from gpaw.basis_data import Basis


class AppelbaumHamann(BaseSetup):
    symbol = 'Si'

    def __init__(self, alpha=0.6102, v1=3.042, v2=-1.372):
        self.alpha = alpha
        self.v1 = v1
        self.v2 = v2

        self.E = 0.0
        self.Z = 14
        self.Nc = 10
        self.Nv = 4
        self.nao = None
        nullspline = Spline(0, 0.5, [0., 0., 0.])
        self.pt_j = [nullspline]
        self.ni = 1
        self.l_j = [0]
        self.f_j = [4]
        self.n_j = [1]
        self.nct = nullspline
        self.tauct = nullspline
        self.Nct = 0.0
        rc = 4.0
        r2_g = np.linspace(0, rc, 100)**2
        x_g = np.exp(-alpha * r2_g)
        self.ghat_l = [Spline(0, rc, 4 * alpha**1.5 / np.pi**0.5 * x_g)]
        self.vbar = Spline(0, rc, 2 * np.pi**0.5 * (v1 + v2 * r2_g) * x_g)
        self.Delta_pL = np.zeros((1, 1))
        self.Delta_iiL = np.zeros((1, 1, 1))
        self.Delta0 = -4 / (4 * np.pi)**0.5
        self.ExxC = 0.0
        self.lmax = 0
        self.K_p = self.M_p = self.MB_p = self.X_p = self.N0_p = np.zeros(1)
        self.M_pp = np.zeros((1, 1))
        self.Kc = 0.0
        self.MB = 0.0
        self.M = 0.0
        self.xc_correction = None
        self.dO_ii = np.zeros((1, 1))
        self.type = 'ah'
        self.fingerprint = None

    def build(self, basis):
        if basis is None:
            basis = Basis('Si', 'sz(dzp)')
        elif isinstance(basis, str):
            basis = Basis('Si', basis)

        self.basis = basis
        self.basis_functions_J = self.basis.tosplines()
        self.nao = self.basis.nao

    def print_info(self, text):
        text('Appelbaum-Hamann pseudo potential')

    def calculate_initial_occupation_numbers(self, magmom, hund, charge,
                                             nspins):
        assert nspins == 1
        return np.array([(2.0, 2.0 / 3, 2.0 / 3, 2.0 / 3)])

    def initialize_density_matrix(self, f_si):
        return np.zeros((len(f_si), 1))

    def get_default_nbands(self):
        return 3



import numpy as np
from ase.units import Hartree
from gpaw.directmin.tools import d_matrix


class KSLCAO:

    """
    This class is used to calculate the gradient of the KS functional
    w.r.t rotation parameters in LCAO
    """

    def __init__(self):
        self.name = 'ks'

    def todict(self):
        return {'name': self.name}

    def get_gradients(self, h_mm, c_nm, f_n, evec, evals, kpt, wfs, timer,
                      matrix_exp, representation, ind_up, constraints):

        with timer('Construct Gradient Matrix'):
            use_egdecomp = matrix_exp == 'egdecomp'
            
            hc_mn, h_ij, h_ia = self.get_ham_in_mol_orb_representation(
                h_mm, c_nm, f_n, representation, use_egdecomp)
            
            with timer('Residual'):
                error = self.get_residual_error(
                    hc_mn, kpt.S_MM, c_nm, h_ij, f_n, wfs.nvalence,
                    constraints)

            if not use_egdecomp:
                if representation == 'u-invar':
                    h_ij = h_ia
                else:
                    if representation == 'full' and wfs.dtype == complex:
                        indl = np.tril_indices(h_ij.shape[0], -1)
                    else:
                        indl = np.tril_indices(h_ij.shape[0])
                    h_ij[indl] = np.inf
                    if representation == 'sparse':
                        h_ij = np.concatenate((h_ij, h_ia), axis=1)
                h_ij = h_ij.ravel()
                h_ij = h_ij[h_ij != np.inf]
        
                ones_mat = np.ones(shape=(len(f_n), len(f_n)))
                anti_occ = f_n * ones_mat - f_n[:, np.newaxis] * ones_mat
                anti_occ = anti_occ[ind_up]
                grad = np.ascontiguousarray(anti_occ * h_ij)

            else:
                h_ij = f_n * h_ij - f_n[:, np.newaxis] * h_ij
                grad = np.ascontiguousarray(h_ij)

                with timer('Use Eigendecomposition'):
                    grad = self.get_exact_gradient_matrix(grad, evec, evals)

                grad = grad[ind_up]

        if wfs.dtype == float:
            grad = grad.real

        if constraints:
            constrain_grad(grad, constraints, ind_up)

        return 2.0 * grad, error

    def get_ham_in_mol_orb_representation(self, h_mm, c_nm, f_n,
                                          representation, full_ham):

        """
        H = (C_nM @ H_MM @ C_nM.T.conj()).conj()
        for sparse and u_inv representation we calculate
        H_ij and H_ia, where i,j -- occupied and a - virtual
        H_ij is really needed only to calculate the residual later
        but it is not needed for u-invar representation.
        When matrix exp calculated using egdecomp method
        we need the whole matrix H though

        :return: H@C_nM[:occ].T, H_ij, H_ia or
                 H@C_nM.T, H, H_ia
        """

        occ = sum(f_n > 1.0e-10)
        if representation in ['sparse', 'u-invar'] \
                and not full_ham:
            hc_mn = h_mm.conj() @ c_nm[:occ].T
            h_ij = hc_mn.T.conj() @ c_nm[:occ].T
            h_ia = hc_mn.T.conj() @ c_nm[occ:].T
        else:
            hc_mn = h_mm.conj() @ c_nm.T
            h_ij = c_nm.conj() @ hc_mn
            h_ia = h_ij[:occ][:, occ:]

        return hc_mn, h_ij, h_ia
    
    def get_residual_error(
            self, hc_mn, S_MM, c_nm, h_ij, f_n, nvalence, constraints=None):
        """
        Calculate residual error of KS equations
        """
        occ = sum(f_n > 1.0e-10)
        hc_mn = hc_mn[:, :occ] - \
            S_MM.conj() @ c_nm[:occ].T @ h_ij[:occ, :occ]
        if constraints:
            # Zero out the components of the residual that are constrained,
            # so that the constrained degrees of freedom are always considered
            # converged
            for con in constraints:
                con1 = con[0]
                hc_mn[:, con1] = 0.0
        norm = sum(hc_mn.conj() * hc_mn * f_n[:occ])
        error = sum(norm.real) * Hartree ** 2 / nvalence
        
        return error

    def get_exact_gradient_matrix(self, h_ij, evec, evals):
 
        """
        Given eigendecomposition of A
        calculate exact gradient matrix
        eq.(14), (39)-(41) from
        arXiv:2101.12597 [physics.comp-ph]
        Comput. Phys. Commun. 267, 108047 (2021).
        :return: gradient matrix
        """

        grad = evec.T.conj() @ h_ij @ evec
        grad = grad * d_matrix(evals)
        grad = evec @ grad @ evec.T.conj()
        for i in range(grad.shape[0]):
            grad[i][i] *= 0.5
        
        return grad


def constrain_grad(grad, constraints, ind_up):
    """
    Zero out the components of the gradient that are constrained, so that no
    optimization step is taken along the constrained degrees of freedom (It
    would be better not to evaluate these components of the gradient to begin
    with.).
    """

    for con in constraints:
        num = -1
        for ind1, ind2 in zip(ind_up[0], ind_up[1]):
            num += 1
            if con == [ind1, ind2]:
                grad[num] = 0.0
    return grad



from ase import Atoms
from ase.calculators.test import numeric_force
from gpaw import GPAW, Mixer, FermiDirac, Davidson
from gpaw.test import equal


def test_generic_H_force():
    a = 4.0
    n = 16
    atoms = Atoms('H',
                  positions=[[1.234, 2.345, 3.456]],
                  cell=(a, a, a),
                  pbc=True)
    calc = GPAW(mode='fd',
                nbands=1,
                gpts=(n, n, n),
                txt=None,
                eigensolver=Davidson(4),
                mixer=Mixer(0.3, 3, 1),
                convergence={'energy': 1e-7},
                occupations=FermiDirac(0.0))
    atoms.calc = calc
    e1 = atoms.get_potential_energy()
    f1 = atoms.get_forces()[0]
    for i in range(3):
        f2i = numeric_force(atoms, 0, i)
        print(f1[i], f2i)
        equal(f1[i], f2i, 0.00025)

    energy_tolerance = 0.001
    force_tolerance = 0.004
    equal(e1, -0.5318, energy_tolerance)
    f1_ref = [-0.29138, -0.3060, -0.3583]
    for i in range(3):
        equal(f1[i], f1_ref[i], force_tolerance)



import numpy as np
from ase.units import Bohr, Hartree

import gpaw.mpi as mpi
from gpaw.tddft.units import eV_to_aufrequency
from gpaw.poisson import PoissonSolver
from gpaw.fd_operators import Gradient
from gpaw.grid_descriptor import GridDescriptor
from gpaw.utilities.extend_grid import extended_grid_descriptor, \
    extend_array, deextend_array, move_atoms


def sendreceive_dict(comm, a_i, dest, b_i, src_i, iitems):
    """
    Send iitems of dictionary b_i distributed according to src_i
    to dictionary a_i in dest.
    """
    requests = []
    for i in iitems:
        # Send and receive
        if comm.rank == dest:
            # Check if dest has it already
            if dest == src_i[i]:
                a_i[i] = b_i[i].copy()
            else:
                requests.append(comm.receive(a_i[i], src_i[i], tag=i,
                                             block=False))
        elif comm.rank == src_i[i]:
            requests.append(comm.send(b_i[i], dest, tag=i, block=False))
    comm.waitall(requests)


class BaseInducedField(object):
    """Partially virtual base class for induced field calculations.

    Attributes:
    -----------
    omega_w: ndarray
        Frequencies for Fourier transform in atomic units
    folding: string
        Folding type ('Gauss' or 'Lorentz' or None)
    width: float
        Width parameter for folding
    Frho_wg: ndarray (complex)
        Fourier transform of induced charge density
    Fphi_wg: ndarray (complex)
        Fourier transform of induced electric potential
    Fef_wvg: ndarray (complex)
        Fourier transform of induced electric field
    Ffe_wg: ndarray (float)
        Fourier transform of field enhancement
    Fbgef_v: ndarray (float)
        Fourier transform of background electric field
    """

    def __init__(self, filename=None, paw=None,
                 frequencies=None, folding='Gauss', width=0.08,
                 readmode=''):
        """
        Parameters:
        -----------
        filename: string
            Filename of a previous ``InducedField`` to be loaded.
            Setting filename disables parameters ``frequencies``,
            ``folding`` and ``width``.
        paw: PAW object
            PAW object for InducedField
        frequencies: ndarray or list of floats
            Frequencies in eV for Fourier transforms.
            This parameter is neglected if ``filename`` is given.
        folding: string
            Folding type: 'Gauss' or 'Lorentz' or None:
            Gaussian or Lorentzian folding or no folding.
            This parameter is neglected if ``filename`` is given.
        width: float
            Width in eV for the Gaussian (sigma) or Lorentzian (eta) folding
            This parameter is neglected if ``filename`` is given.
        """
        self.dtype = complex

        # These are allocated when calculated
        self.fieldgd = None
        self.Frho_wg = None
        self.Fphi_wg = None
        self.Fef_wvg = None
        self.Ffe_wg = None
        self.Fbgef_v = None

        # has variables
        self.has_paw = False
        self.has_field = False

        if not hasattr(self, 'readwritemode_str_to_list'):
            self.readwritemode_str_to_list = \
                {'': ['Frho', 'atoms'],
                 'all': ['Frho', 'Fphi', 'Fef', 'Ffe', 'atoms'],
                 'field': ['Frho', 'Fphi', 'Fef', 'Ffe', 'atoms']}

        if filename is not None:
            self.initialize(paw, allocate=False)
            self.read(filename, mode=readmode)
            return

        self.folding = folding
        self.width = width * eV_to_aufrequency
        self.set_folding(folding, width * eV_to_aufrequency)

        self.omega_w = np.asarray(frequencies) * eV_to_aufrequency
        self.nw = len(self.omega_w)

        self.nv = 3  # dimensionality of the space

        self.initialize(paw, allocate=True)

    def initialize(self, paw, allocate=True):
        self.allocated = False
        self.has_paw = paw is not None

        if self.has_paw:
            # If no paw is given, then the variables
            # marked with "# !" are created in _read().
            # Other variables are not accessible without
            # paw (TODO: could be implemented...).
            self.paw = paw
            self.world = paw.wfs.world                            # !
            self.domain_comm = paw.wfs.gd.comm                    # !
            self.band_comm = paw.wfs.bd.comm                      # !
            self.kpt_comm = paw.wfs.kd.comm                       # !
            self.rank_a = paw.wfs.atom_partition.rank_a
            self.nspins = paw.density.nspins                      # !
            self.setups = paw.wfs.setups
            self.density = paw.density
            self.atoms = paw.atoms                                # !
            self.na = len(self.atoms.get_atomic_numbers())        # !
            self.gd = self.density.gd                             # !
            self.stencil = self.density.stencil
            self.log = paw.log

        if allocate:
            self.allocate()

    def allocate(self):
        self.allocated = True

    def deallocate(self):
        self.fieldgd = None
        self.Frho_wg = None
        self.Fphi_wg = None
        self.Fef_wvg = None
        self.Ffe_wg = None
        self.Fbgef_v = None

    def set_folding(self, folding, width):
        """
        width: float
            Width in atomic units
        """
        if width is None:
            folding = None

        self.folding = folding
        if self.folding is None:
            self.width = None
        else:
            self.width = width

    def get_induced_density(self, from_density, gridrefinement):
        raise RuntimeError('Virtual member function called')

    def calculate_induced_field(self, from_density='comp',
                                gridrefinement=2,
                                extend_N_cd=None,
                                deextend=False,
                                poissonsolver=None,
                                gradient_n=3):
        if self.has_field and \
           from_density == self.field_from_density and \
           self.Frho_wg is not None:
            Frho_wg = self.Frho_wg
            gd = self.fieldgd
        else:
            Frho_wg, gd = self.get_induced_density(from_density,
                                                   gridrefinement)

        # Always extend a bit to get field without jumps
        if extend_N_cd is None:
            extend_N = max(8, 2**int(np.ceil(np.log(gradient_n) / np.log(2.))))
            extend_N_cd = extend_N * np.ones(shape=(3, 2), dtype=int)
            deextend = True

        # Extend grid
        oldgd = gd
        egd, cell_cv, move_c = \
            extended_grid_descriptor(gd, extend_N_cd=extend_N_cd)
        Frho_we = egd.zeros((self.nw,), dtype=self.dtype)
        for w in range(self.nw):
            extend_array(gd, egd, Frho_wg[w], Frho_we[w])
        Frho_wg = Frho_we
        gd = egd
        if not deextend:
            # TODO: this will make atoms unusable with original grid
            self.atoms.set_cell(cell_cv, scale_atoms=False)
            move_atoms(self.atoms, move_c)

        # Allocate arrays
        Fphi_wg = gd.zeros((self.nw,), dtype=self.dtype)
        Fef_wvg = gd.zeros((self.nw, self.nv,), dtype=self.dtype)
        Ffe_wg = gd.zeros((self.nw,), dtype=float)

        # Poissonsolver
        if poissonsolver is None:
            poissonsolver = PoissonSolver(name='fd', eps=1e-20)
        poissonsolver.set_grid_descriptor(gd)

        for w in range(self.nw):
            # TODO: better output of progress
            # parprint('%d' % w)
            calculate_field(gd, Frho_wg[w], self.Fbgef_v,
                            Fphi_wg[w], Fef_wvg[w], Ffe_wg[w],
                            poissonsolver=poissonsolver,
                            nv=self.nv,
                            gradient_n=gradient_n)

        # De-extend grid
        if deextend:
            Frho_wo = oldgd.zeros((self.nw,), dtype=self.dtype)
            Fphi_wo = oldgd.zeros((self.nw,), dtype=self.dtype)
            Fef_wvo = oldgd.zeros((self.nw, self.nv,), dtype=self.dtype)
            Ffe_wo = oldgd.zeros((self.nw,), dtype=float)
            for w in range(self.nw):
                deextend_array(oldgd, gd, Frho_wo[w], Frho_wg[w])
                deextend_array(oldgd, gd, Fphi_wo[w], Fphi_wg[w])
                deextend_array(oldgd, gd, Ffe_wo[w], Ffe_wg[w])
                for v in range(self.nv):
                    deextend_array(oldgd, gd, Fef_wvo[w][v], Fef_wvg[w][v])
            Frho_wg = Frho_wo
            Fphi_wg = Fphi_wo
            Fef_wvg = Fef_wvo
            Ffe_wg = Ffe_wo
            gd = oldgd

        # Store results
        self.has_field = True
        self.field_from_density = from_density
        self.fieldgd = gd
        self.Frho_wg = Frho_wg
        self.Fphi_wg = Fphi_wg
        self.Fef_wvg = Fef_wvg
        self.Ffe_wg = Ffe_wg

    def _parse_readwritemode(self, mode):
        if isinstance(mode, str):
            try:
                readwrites = self.readwritemode_str_to_list[mode]
            except KeyError:
                raise IOError('unknown readwrite mode string')
        elif isinstance(mode, list):
            readwrites = mode
        else:
            raise IOError('unknown readwrite mode type')

        if any(k in readwrites for k in ['Frho', 'Fphi', 'Fef', 'Ffe']):
            readwrites.append('field')

        return readwrites

    def read(self, filename, mode='', idiotproof=True):
        if idiotproof and not filename.endswith('.ind'):
            raise IOError('Filename must end with `.ind`.')

        reads = self._parse_readwritemode(mode)

        # Open reader
        from gpaw.io import Reader
        reader = Reader(filename)
        self._read(reader, reads)
        reader.close()
        self.world.barrier()

    def _read(self, reader, reads):
        r = reader

        # Test data type
        dtype = {'float': float, 'complex': complex}[r.dtype]
        if dtype != self.dtype:
            raise IOError('Data is an incompatible type.')

        # Read dimensions
        na = r.na
        self.nv = r.nv
        nspins = r.nspins
        ng = r.ng

        # Background electric field
        self.Fbgef_v = r.Fbgef_v

        if self.has_paw:
            # Test dimensions
            if na != self.na:
                raise IOError('natoms is incompatible with calculator')
            if nspins != self.nspins:
                raise IOError('nspins is incompatible with calculator')
            if (ng != self.gd.get_size_of_global_array()).any():
                raise IOError('grid is incompatible with calculator')
            if (self.Fbgef_v != self.paw.kick_strength).any():
                raise IOError('kick is incompatible with calculator')
        else:
            # Construct objects / assign values without paw
            self.na = na
            self.nspins = nspins

            from ase.io.trajectory import read_atoms
            self.atoms = read_atoms(r.atoms)

            self.world = mpi.world
            self.gd = GridDescriptor(ng + 1, self.atoms.get_cell() / Bohr,
                                     pbc_c=False, comm=self.world)
            self.domain_comm = self.gd.comm
            self.band_comm = mpi.SerialCommunicator()
            self.kpt_comm = mpi.SerialCommunicator()

        # Folding
        folding = r.folding
        width = r.width
        self.set_folding(folding, width)

        # Frequencies
        self.omega_w = r.omega_w
        self.nw = len(self.omega_w)

        # Read field
        if 'field' in reads:
            nfieldg = r.nfieldg
            self.has_field = True
            self.field_from_density = r.field_from_density
            self.fieldgd = self.gd.new_descriptor(N_c=nfieldg + 1)

            def readarray(name, shape, dtype):
                if name.split('_')[0] in reads:
                    setattr(self, name, self.fieldgd.empty(shape, dtype=dtype))
                    self.fieldgd.distribute(r.get(name), getattr(self, name))

            readarray('Frho_wg', (self.nw,), self.dtype)
            readarray('Fphi_wg', (self.nw,), self.dtype)
            readarray('Fef_wvg', (self.nw, self.nv), self.dtype)
            readarray('Ffe_wg', (self.nw,), float)

    def write(self, filename, mode='', idiotproof=True):
        """
        Parameters
        ----------
        mode: string or list of strings


        """
        if idiotproof and not filename.endswith('.ind'):
            raise IOError('Filename must end with `.ind`.')

        writes = self._parse_readwritemode(mode)

        if 'field' in writes and self.fieldgd is None:
            raise IOError('field variables cannot be written ' +
                          'before they are calculated')

        from gpaw.io import Writer
        writer = Writer(filename, self.world, tag='INDUCEDFIELD')
        # Actual write
        self._write(writer, writes)
        # Make sure slaves don't return before master is done
        writer.close()
        self.world.barrier()

    def _write(self, writer, writes):
        # Write parameters/dimensions
        writer.write(dtype={float: 'float', complex: 'complex'}[self.dtype],
                     folding=self.folding,
                     width=self.width,
                     na=self.na,
                     nv=self.nv,
                     nspins=self.nspins,
                     ng=self.gd.get_size_of_global_array())

        # Write field grid
        if 'field' in writes:
            writer.write(field_from_density=self.field_from_density,
                         nfieldg=self.fieldgd.get_size_of_global_array())

        # Write frequencies
        writer.write(omega_w=self.omega_w)

        # Write background electric field
        writer.write(Fbgef_v=self.Fbgef_v)

        from ase.io.trajectory import write_atoms
        write_atoms(writer.child('atoms'), self.atoms)

        if 'field' in writes:
            def writearray(name, shape, dtype):
                if name.split('_')[0] in writes:
                    writer.add_array(name, shape, dtype)
                a_wxg = getattr(self, name)
                for w in range(self.nw):
                    writer.fill(self.fieldgd.collect(a_wxg[w]))

            shape = (self.nw,) + tuple(self.fieldgd.get_size_of_global_array())
            writearray('Frho_wg', shape, self.dtype)
            writearray('Fphi_wg', shape, self.dtype)
            writearray('Ffe_wg', shape, float)

            shape3 = shape[:1] + (self.nv,) + shape[1:]
            writearray('Fef_wvg', shape3, self.dtype)


def calculate_field(gd, rho_g, bgef_v,
                    phi_g, ef_vg, fe_g,  # preallocated numpy arrays
                    poissonsolver,
                    nv=3,
                    gradient_n=3):

    dtype = rho_g.dtype
    yes_complex = dtype == complex

    phi_g[:] = 0.0
    ef_vg[:] = 0.0
    fe_g[:] = 0.0
    tmp_g = gd.zeros(dtype=float)

    # Potential, real part
    poissonsolver.solve(tmp_g, rho_g.real.copy())
    phi_g += tmp_g
    # Potential, imag part
    if yes_complex:
        tmp_g[:] = 0.0
        poissonsolver.solve(tmp_g, rho_g.imag.copy())
        phi_g += 1.0j * tmp_g

    # Gradient
    gradient = [Gradient(gd, v, scale=1.0, n=gradient_n)
                for v in range(nv)]
    for v in range(nv):
        # Electric field, real part
        gradient[v].apply(-phi_g.real, tmp_g)
        ef_vg[v] += tmp_g
        # Electric field, imag part
        if yes_complex:
            gradient[v].apply(-phi_g.imag, tmp_g)
            ef_vg[v] += 1.0j * tmp_g

    # Electric field enhancement
    tmp_g[:] = 0.0  # total electric field norm
    bgefnorm = 0.0  # background electric field norm
    for v in range(nv):
        tmp_g += np.absolute(bgef_v[v] + ef_vg[v])**2
        bgefnorm += np.absolute(bgef_v[v])**2

    tmp_g = np.sqrt(tmp_g)
    bgefnorm = np.sqrt(bgefnorm)

    fe_g[:] = tmp_g / bgefnorm


def zero_pad(a):
    """Add zeros to both sides of all axis on array a.

    Not parallel safe.
    """

    z_shape = np.zeros(shape=len(a.shape))
    z_shape[-3:] = 2
    b_shape = np.array(a.shape) + z_shape

    b = np.zeros(shape=b_shape, dtype=a.dtype)

    b[..., 1:-1, 1:-1, 1:-1] = a
    return b


# TOOD: remove/edit this function
def calculate_oscstr(Fn_wg, omega_w, box, kick):
    omega_w = np.array(omega_w)
    omega_w *= eV_to_aufrequency  # to a.u.
    box = box.copy() / Bohr  # to a.u
    volume = box.prod()
    ng = np.array(Fn_wg[0].shape)
    dV = volume / (ng - 1).prod()  # Note: data is zeropadded to both sides

    Fn_wg_im = Fn_wg.imag
    if kick[0] != 0.0:
        osc_w = -2 * omega_w / np.pi / kick[0] * \
            ((Fn_wg_im.sum(axis=2)).sum(axis=2) *
             np.linspace(0, box[0], ng[0])).sum(axis=1) * dV
    elif kick[1] != 0.0:
        osc_w = -2 * omega_w / np.pi / kick[1] * \
            ((Fn_wg_im.sum(axis=1)).sum(axis=2) *
             np.linspace(0, box[1], ng[1])).sum(axis=1) * dV
    elif kick[2] != 0.0:
        osc_w = -2 * omega_w / np.pi / kick[2] * \
            ((Fn_wg_im.sum(axis=1)).sum(axis=1) *
             np.linspace(0, box[2], ng[2])).sum(axis=1) * dV
    return osc_w / Hartree  # to 1/eV


# TOOD: remove/edit this function
def calculate_polarizability(Fn_wg, box, kick):
    box = box.copy() / Bohr  # to a.u
    volume = box.prod()
    ng = np.array(Fn_wg[0].shape)
    dV = volume / (ng - 1).prod()  # Note: data is zeropadded to both sides

    if kick[0] != 0.0:
        pol_w = -1.0 / kick[0] * \
            ((Fn_wg.sum(axis=2)).sum(axis=2) *
             np.linspace(0, box[0], ng[0])).sum(axis=1) * dV
    elif kick[1] != 0.0:
        pol_w = -1.0 / kick[1] * \
            ((Fn_wg.sum(axis=1)).sum(axis=2) *
             np.linspace(0, box[1], ng[1])).sum(axis=1) * dV
    elif kick[2] != 0.0:
        pol_w = -1.0 / kick[2] * \
            ((Fn_wg.sum(axis=1)).sum(axis=1) *
             np.linspace(0, box[2], ng[2])).sum(axis=1) * dV
    return pol_w / Hartree**2  # to 1/eV**2



"""Test to make sure spin-polarized GW calculations don't break."""

# General modules
import pytest
from gpaw.response.g0w0 import G0W0
from gpaw.mpi import world


@pytest.mark.response
def test_gw_spinpol(in_tmp_dir, gpw_files, needs_ase_master):

    if world.size > 1:
        nblocks = 2
    else:
        nblocks = 1

    gw = G0W0(gpw_files['h2_bcc_afm'],
              nbands=4,  # keep consistent with gpw nbands
              ecut=100,
              kpts=[(0, 0, 0)],
              nblocks=nblocks,
              relbands=(-1, 1))
    result = gw.calculate()

    # Make sure gaps in both spin-channels are the same and don't change.
    # test values do not necessarily reflect those in literature. They simply
    # ensure the value does not change.
    lda_sn = result['eps'][:, 0]
    lda_gap_s = lda_sn[:, 1] - lda_sn[:, 0]
    assert abs(lda_gap_s - 10.5095).max() < 0.01
    qp_sn = result['qp'][:, 0]
    qp_gap_s = qp_sn[:, 1] - qp_sn[:, 0]
    assert abs(qp_gap_s - 9.8156).max() < 0.01



from functools import partial

import pytest
from ase.units import Ha
import numpy as np

from gpaw.occupations import (fermi_dirac, marzari_vanderbilt,
                              methfessel_paxton, FermiDiracCalculator,
                              ZeroWidth)
from gpaw.tetrahedron import TetrahedronMethod


funcs = []
for w in [0.1, 0.5]:
    funcs.append(partial(fermi_dirac, fermi_level=0.0, width=w))
    funcs.append(partial(marzari_vanderbilt, fermi_level=0.0, width=w))
    for n in range(4):
        funcs.append(partial(methfessel_paxton,
                             fermi_level=0.0, order=n, width=w))


@pytest.mark.parametrize('func', funcs)
@pytest.mark.ci
def test_occupations(func):
    for e in [-0.3 / Ha, 0, 0.1 / Ha, 1.2 / Ha]:
        n0, d0, S0 = func(e)
        x = 0.000001
        fp, dp, Sp = func(e + x)
        fm, dm, Sm = func(e - x)
        d = -(fp - fm) / (2 * x)
        dS = Sm - Sp
        dn = fp - fm
        print(d - d0, dS - e * dn)
        assert abs(d - d0) < 3e-5
        assert abs(dS - e * dn) < 1e-13


@pytest.mark.parametrize(
    'e_kn,w_k,ne',
    [([[0.0, 1.0]], [1.0], 2),
     ([[0.0, 1.0], [0.0, 2.0]], [0.5, 0.5], 1.5),
     ([[0.0, 1.0, 2.0], [0.0, 2.0, 2.0]], [0.5, 0.5], 2)])
@pytest.mark.ci
def test_occ_obj(e_kn, w_k, ne):
    print('*' * 42)
    print(e_kn)
    for occ in [FermiDiracCalculator(0.1),
                ZeroWidth(),
                TetrahedronMethod(np.diag([2, 1, 1]), (len(w_k), 1, 1))]:
        f_kn, fl, s = occ.calculate(ne, np.array(e_kn), w_k)
        print(occ)
        print(f_kn)
        print(fl, s, (w_k @ f_kn).sum())
        assert (w_k @ f_kn).sum() == pytest.approx(ne, abs=1e-14)



from __future__ import annotations

from typing import Callable

import numpy as np
from gpaw.core import UGArray, UGDesc
from gpaw.gpu import as_xp
from gpaw.gpu import cupy as cp
from gpaw.new.calculation import DFTState
from gpaw.typing import Array2D, Array4D
from gpaw.xc import XC
from gpaw.xc.functional import XCFunctional as OldXCFunctional
from gpaw.xc.gga import add_gradient_correction, gga_vars
from gpaw.xc.mgga import MGGA
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new import zips
from gpaw.new.pwfd.wave_functions import PWFDWaveFunctions


def create_functional(xc: OldXCFunctional | str | dict,
                      grid: UGDesc) -> Functional:
    if isinstance(xc, (str, dict)):
        xc = XC(xc)
    if xc.type == 'MGGA':
        return MGGAFunctional(xc, grid)
    assert xc.type in {'LDA', 'GGA'}, xc
    return LDAOrGGAFunctional(xc, grid)


class Functional:
    def __init__(self,
                 xc: OldXCFunctional,
                 grid: UGDesc):
        self.xc = xc
        self.grid = grid
        self.setup_name = self.xc.get_setup_name()
        self.name = self.xc.name
        self.type = self.xc.type
        self.xc.set_grid_descriptor(grid._gd)

    def __str__(self):
        return f'name: {self.xc.get_description()}'

    def calculate_paw_correction(self, setup, d, h=None):
        return self.xc.calculate_paw_correction(setup, d, h)

    def get_setup_name(self) -> str:
        return self.name

    def stress_contribution(self,
                            state: DFTState,
                            interpolate: Callable[[UGArray], UGArray]
                            ) -> Array2D:
        raise NotImplementedError


class LDAOrGGAFunctional(Functional):
    def calculate(self,
                  nt_sr: UGArray,
                  taut_sr: UGArray | None = None) -> tuple[float,
                                                           UGArray,
                                                           UGArray | None]:
        xp = nt_sr.xp
        vxct_sr = nt_sr.new()
        if xp is np:
            vxct_sr.data[:] = 0.0
            exc = self.xc.calculate(self.xc.gd, nt_sr.data, vxct_sr.data)
        else:
            vxct_np_sr = np.zeros(nt_sr.data.shape)
            assert isinstance(nt_sr.data, cp.ndarray)
            exc = self.xc.calculate(nt_sr.desc._gd, nt_sr.data.get(),
                                    vxct_np_sr)
            vxct_sr.data[:] = xp.asarray(vxct_np_sr)
        return exc, vxct_sr, None

    def stress_contribution(self,
                            state: DFTState,
                            interpolate: Callable[[UGArray], UGArray]
                            ) -> Array2D:
        ibzwfs = state.ibzwfs
        if ibzwfs.kpt_comm.rank == 0 and ibzwfs.band_comm.rank == 0:
            nt_sr = interpolate(state.density.nt_sR)
            s_vv = self.xc.stress_tensor_contribution(as_xp(nt_sr.data, np),
                                                      skip_sum=True)
            return as_xp(s_vv, nt_sr.xp)
        return state.density.nt_sR.xp.zeros((3, 3))


class MGGAFunctional(Functional):
    def get_setup_name(self):
        return 'PBE'

    def calculate(self,
                  nt_sr,
                  taut_sr) -> tuple[float, UGArray, UGArray | None]:
        gd = self.xc.gd
        assert isinstance(self.xc, MGGA), self.xc
        sigma_xr, dedsigma_xr, gradn_svr = gga_vars(gd, self.xc.grad_v,
                                                    nt_sr.data)
        e_r = self.grid.empty()
        if taut_sr is None:
            taut_sr = nt_sr.new(zeroed=True)
        dedtaut_sr = taut_sr.new()
        vxct_sr = taut_sr.new()
        vxct_sr.data[:] = 0.0
        self.xc.kernel.calculate(e_r.data, nt_sr.data, vxct_sr.data,
                                 sigma_xr, dedsigma_xr,
                                 taut_sr.data, dedtaut_sr.data)
        add_gradient_correction(self.xc.grad_v, gradn_svr, sigma_xr,
                                dedsigma_xr, vxct_sr.data)
        return e_r.integrate(), vxct_sr, dedtaut_sr

    def stress_contribution(self,
                            state: DFTState,
                            interpolate: Callable[[UGArray], UGArray]
                            ) -> Array2D:
        stress_vv = np.zeros((3, 3))

        taut_swR = _taut(state.ibzwfs, state.density.nt_sR.desc)
        if taut_swR is None:
            return stress_vv

        dedtaut_sR = state.potential.dedtaut_sR
        assert dedtaut_sR is not None
        dedtaut_sr = interpolate(dedtaut_sR)
        for taut_wR, dedtaut_r in zips(taut_swR, dedtaut_sr):
            w = 0
            for v1 in range(3):
                for v2 in range(v1, 3):
                    taut_r = interpolate(taut_wR[w])
                    s = taut_r.integrate(dedtaut_r, skip_sum=True)
                    stress_vv[v1, v2] -= s
                    if v2 != v1:
                        stress_vv[v2, v1] -= s
                    w += 1

        nt_sr = interpolate(state.density.nt_sR)
        taut_sR = state.density.taut_sR
        assert taut_sR is not None
        taut_sr = interpolate(taut_sR)
        assert isinstance(self.xc, MGGA)
        stress_vv += _mgga(self.xc, nt_sr.data, taut_sr.data)

        return stress_vv


def _taut(ibzwfs: IBZWaveFunctions, grid: UGDesc) -> UGArray | None:
    """Calculate upper half of symmetric ked tensor.

    Returns ``taut_swR=taut_svvR``.  Mapping from ``w`` to ``vv``::

        0 1 2
        . 3 4
        . . 5

    Only cores with ``kpt_comm.rank==0`` and ``band_comm.rank==0``
    return the uniform grids - other cores return None.
    """
    # "1" refers to undistributed arrays
    dpsit1_vR = grid.new(comm=None, dtype=ibzwfs.dtype).empty(3)
    taut1_swR = grid.new(comm=None).zeros((ibzwfs.nspins, 6))
    assert isinstance(taut1_swR, UGArray)  # Argggghhh!
    domain_comm = grid.comm

    for wfs in ibzwfs:
        assert isinstance(wfs, PWFDWaveFunctions)
        psit_nG = wfs.psit_nX
        pw = psit_nG.desc

        pw1 = pw.new(comm=None)
        psit1_G = pw1.empty()
        iGpsit1_G = pw1.empty()
        Gplusk1_Gv = pw1.reciprocal_vectors()

        occ_n = wfs.weight * wfs.spin_degeneracy * wfs.myocc_n

        (N,) = psit_nG.mydims
        for n1 in range(0, N, domain_comm.size):
            n2 = min(n1 + domain_comm.size, N)
            psit_nG[n1:n2].gather_all(psit1_G)
            n = n1 + domain_comm.rank
            if n >= N:
                continue
            f = occ_n[n]
            if f == 0.0:
                continue
            for Gplusk1_G, dpsit1_R in zips(Gplusk1_Gv.T, dpsit1_vR):
                iGpsit1_G.data[:] = psit1_G.data
                iGpsit1_G.data *= 1j * Gplusk1_G
                iGpsit1_G.ifft(out=dpsit1_R)
            w = 0
            for v1 in range(3):
                for v2 in range(v1, 3):
                    taut1_swR[wfs.spin, w].data += (
                        f * (dpsit1_vR[v1].data.conj() *
                             dpsit1_vR[v2].data).real)
                    w += 1

    ibzwfs.kpt_comm.sum(taut1_swR.data, 0)
    if ibzwfs.kpt_comm.rank == 0:
        ibzwfs.band_comm.sum(taut1_swR.data, 0)
        if ibzwfs.band_comm.rank == 0:
            domain_comm.sum(taut1_swR.data, 0)
            if domain_comm.rank == 0:
                symmetries = ibzwfs.ibz.symmetries
                taut1_swR.symmetrize(symmetries.rotation_scc,
                                     symmetries.translation_sc)
            taut_swR = grid.empty((ibzwfs.nspins, 6))
            taut_swR.scatter_from(taut1_swR)
            return taut_swR
    return None


def _mgga(xc: MGGA, nt_sr: Array4D, taut_sr: Array4D) -> Array2D:
    # The GGA and LDA part of this should be factored out!
    sigma_xr, dedsigma_xr, gradn_svr = gga_vars(xc.gd,
                                                xc.grad_v,
                                                nt_sr)
    nspins = len(nt_sr)
    dedtaut_sr = np.empty_like(nt_sr)
    vt_sr = xc.gd.zeros(nspins)
    e_r = xc.gd.empty()
    xc.kernel.calculate(e_r, nt_sr, vt_sr, sigma_xr, dedsigma_xr,
                        taut_sr, dedtaut_sr)

    def integrate(a1_r, a2_r=None):
        return xc.gd.integrate(a1_r, a2_r, global_integral=False)

    P = integrate(e_r)
    for vt_r, nt_r in zip(vt_sr, nt_sr):
        P -= integrate(vt_r, nt_r)
    for sigma_r, dedsigma_r in zip(sigma_xr, dedsigma_xr):
        P -= 2 * integrate(sigma_r, dedsigma_r)
    for taut_r, dedtaut_r in zip(taut_sr, dedtaut_sr):
        P -= integrate(taut_r, dedtaut_r)

    stress_vv = P * np.eye(3)
    for v1 in range(3):
        for v2 in range(3):
            stress_vv[v1, v2] -= integrate(gradn_svr[0, v1] *
                                           gradn_svr[0, v2],
                                           dedsigma_xr[0]) * 2
            if nspins == 2:
                stress_vv[v1, v2] -= integrate(gradn_svr[0, v1] *
                                               gradn_svr[1, v2],
                                               dedsigma_xr[1]) * 2
                stress_vv[v1, v2] -= integrate(gradn_svr[1, v1] *
                                               gradn_svr[1, v2],
                                               dedsigma_xr[2]) * 2
    return stress_vv



import os
import time

import numpy as np

import ase.db
from ase.units import Rydberg
from ase.utils import opencew
from ase.calculators.calculator import kpts2mp
from ase.io import Trajectory
from ase.calculators.abinit import Abinit
from ase.test.tasks.dcdft import DeltaCodesDFTCollection as Collection

c = ase.db.connect('dcdft_abinit_fhi.db')

ecut = 120

kptdensity = 16.0
width = 0.01

linspace = (0.98, 1.02, 5)  # eos numpy's linspace
linspacestr = ''.join([str(t) + 'x' for t in linspace])[:-1]

code = 'abinit' + '-' + '_c' + str(ecut) + '_e' + linspacestr
code = code + '_k' + str(kptdensity) + '_w' + str(width)

collection = Collection()

for name in collection.names:
    # save all steps in one traj file in addition to the database
    # we should only used the database c.reserve, but here
    # traj file is used as another lock ...
    fd = opencew(name + '_' + code + '.traj')
    if fd is None:
        continue
    traj = Trajectory(name + '_' + code + '.traj', 'w')
    atoms = collection[name]
    if name == 'Mn':  # fails to find the right magnetic state
        atoms.set_initial_magnetic_moments([10., 20., -10., -20.])
    if name == 'Co':  # fails to find the right magnetic state
        atoms.set_initial_magnetic_moments([10., 10.])
    if name == 'Ni':  # fails to find the right magnetic state
        atoms.set_initial_magnetic_moments([10., 10., 10., 10.])
    cell = atoms.get_cell()
    kpts = tuple(kpts2mp(atoms, kptdensity, even=True))
    kwargs = {}
    # loop over EOS linspace
    for n, x in enumerate(np.linspace(linspace[0], linspace[1], linspace[2])):
        id = c.reserve(name=name, ecut=ecut,
                       linspacestr=linspacestr,
                       kptdensity=kptdensity, width=width,
                       x=x)
        if id is None:
            continue
        # perform EOS step
        atoms.set_cell(cell * x, scale_atoms=True)
        # set calculator
        atoms.calc = Abinit(
            label=name + '_' + code + '_' + str(n),
            xc='PBE',
            kpts=kpts,
            ecut=ecut * Rydberg,
            occopt=3,
            tsmear=width,
            toldfe=1.0e-6,
            nstep=900,
            diemix=0.1,
            fband=0.95,
            # http://forum.abinit.org/viewtopic.php?f=8&t=35
            chksymbreak=0,
            nsym=1,  # various symmetry problems with various abinits ...
        )
        atoms.calc.set(**kwargs)  # remaining calc keywords
        t = time.time()
        atoms.get_potential_energy()
        c.write(atoms,
                name=name, ecut=ecut,
                linspacestr=linspacestr,
                kptdensity=kptdensity, width=width,
                x=x,
                time=time.time() - t)
        traj.write(atoms)
        wfk = name + '_' + code + '_' + str(n) + 'o_WFK'
        if os.path.exists(wfk):
            os.remove(wfk)
        del c[id]



import pytest
import gpaw.gpu


@pytest.fixture(scope='session')
def gpu():
    try:
        gpaw.gpu.setup()
    except AttributeError:
        pytest.skip('Not compiled with gpu=True')
    if gpaw.gpu.cupy_is_fake:
        pytest.skip('No cupy')
